import pytest
import jax
import jaxlib
import jax.numpy as np
import tensorflow as tf
from test.aggregated_defaults import aggregatedTests
from test.gradient_defaults import gradientTests
from imnn.imnn import DatasetGradientIMNN


class datasetGradientTests(
        aggregatedTests, gradientTests):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.kwargs.pop("fiducial")
        self.kwargs.pop("derivative")
        self.kwargs.pop("validation_fiducial")
        self.kwargs.pop("validation_derivative")
        self.reduced_kwargs.pop("fiducial")
        self.reduced_kwargs.pop("derivative")
        self.reduced_kwargs.pop("validation_fiducial")
        self.reduced_kwargs.pop("validation_derivative")

        self.kwargs["main"] = (
            self.fiducial[:self.n_d].reshape(
                (self.n_devices,
                 self.n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape),
            self.derivative.reshape(
                (self.n_devices,
                 self.n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape + (self.n_params,)))

        self.kwargs["validation_main"] = (
            self.validation_fiducial[:self.n_d].reshape(
                (self.n_devices,
                 self.n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape),
            self.validation_derivative.reshape(
                (self.n_devices,
                 self.n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape + (self.n_params,)))

        self.kwargs["remaining"] = self.fiducial[self.n_d:].reshape(
            (self.n_devices,
             (self.n_s - self.n_d) // (self.n_devices * self.n_per_device),
             self.n_per_device)
            + self.input_shape)

        self.kwargs["validation_remaining"] = \
            self.validation_fiducial[self.n_d:].reshape(
                (self.n_devices,
                 (self.n_s - self.n_d) // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.reduced_kwargs["main"] = (
            self.fiducial[:self.reduced_n_d].reshape(
                (self.n_devices,
                 self.reduced_n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape),
            self.reduced_derivative.reshape(
                (self.n_devices,
                 self.reduced_n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape + (self.n_params,)))

        self.reduced_kwargs["validation_main"] = (
            self.validation_fiducial[:self.reduced_n_d].reshape(
                (self.n_devices,
                 self.reduced_n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape),
            self.reduced_validation_derivative.reshape(
                (self.n_devices,
                 self.reduced_n_d // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape + (self.n_params,)))

        self.reduced_kwargs["remaining"] = \
            self.fiducial[self.reduced_n_d:].reshape(
                (self.n_devices,
                 (self.n_s - self.reduced_n_d) //
                    (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.reduced_kwargs["validation_remaining"] = \
            self.validation_fiducial[self.reduced_n_d:].reshape(
                (self.n_devices,
                 (self.n_s - self.reduced_n_d) //
                    (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.arrays = ["θ_fid"]

    def preload(self, dictionary, state=False, validate=False):
        dictionary["host"] = jax.devices("cpu")[0]
        dictionary["devices"] = jax.devices()
        dictionary["main"] = [
            tf.data.Dataset.zip(
                (tf.data.Dataset.from_tensor_slices(fid),
                 tf.data.Dataset.from_tensor_slices(der))
            ).repeat().as_numpy_iterator()
            for fid, der in zip(*dictionary["main"])]
        dictionary["remaining"] = [
            tf.data.Dataset.from_tensor_slices(
                fid).repeat().as_numpy_iterator()
            for fid in dictionary["remaining"]]
        if state:
            dictionary["key_or_state"] = self.state
        if (not self.simulate) and (not validate):
            dictionary.pop("validation_main")
            dictionary.pop("validation_remaining")
        else:
            dictionary["validation_main"] = [
                tf.data.Dataset.zip(
                    (tf.data.Dataset.from_tensor_slices(fid),
                     tf.data.Dataset.from_tensor_slices(der))
                ).repeat().as_numpy_iterator()
                for fid, der in zip(*dictionary["validation_main"])]
            dictionary["validation_remaining"] = [
                tf.data.Dataset.from_tensor_slices(
                    fid).repeat().as_numpy_iterator()
                for fid in dictionary["validation_remaining"]]

        return dictionary

    def specific_exceptions(self, variable, input_variable, kwargs):
        if variable == "validation_main":
            if "validation_main" not in kwargs.keys():
                return True
        if variable == "validation_remaining":
            if "validation_remaining" not in kwargs.keys():
                return True
        if variable == "devices":
            if input_variable is list():
                if len(input_variable) < 1:
                    kwargs[variable] = input_variable
                    with pytest.raises(ValueError) as info:
                        self.imnn(**kwargs)
                    assert info.match("`devices` has no elements in")
                    return True
                if not all(
                        [isinstance(device, jaxlib.xla_extension.Device)
                         for device in input_variable]):
                    kwargs[variable] = input_variable
                    with pytest.raises(TypeError) as info:
                        self.imnn(**kwargs)
                    assert info.match(
                        "`all elements of `devices` must be xla devices")
                    return True
        if variable == "host":
            if input_variable is None:
                return False
            if not isinstance(input_variable, jaxlib.xla_extension.Device):
                kwargs[variable] = input_variable
                with pytest.raises(TypeError) as info:
                    self.imnn(**kwargs)
                assert info.match(
                    "`host` must be an xla device but is a "
                    f"{type(input_variable)}")
                return True
        return False

    def splitting_resize(self, _kwargs):
        _kwargs["main"] = None
        _kwargs["remaining"] = None
        if "validation_main" in _kwargs.keys():
            _kwargs["validation_main"] = None
            _kwargs["validation_remaining"] = None
        return _kwargs


test = datasetGradientTests(
    imnn=DatasetGradientIMNN,
    filename="dataset_gradient",
    n_per_device=100)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable",
    [None, list(), 1., 1, np.zeros((1,)), test.rng, tuple(), (0, 0),
     (test.model[0], 0), test.bad_model, test.state])
@pytest.mark.parametrize("variable", test.kwargs.keys())
def test_initialisation_parameters_(
        variable, kwargs, input_variable, state, validate):
    test.initialise_parameters(
        variable, kwargs, input_variable, state=state, validate=validate)


@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("state", [False, True])
@pytest.mark.parametrize("variable", ["n_s", "n_d", "same"])
def test_splitting_(variable, validate, state):
    test.splitting(variable, test.kwargs, state=state, validate=validate)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable", [None, list(), 1., 1, np.zeros((1,)), test.rng])
@pytest.mark.parametrize("variable", test.fit_kwargs.keys())
def test_fit_parameters_(variable, kwargs, input_variable, state, validate):
    test.fit_parameters(
        variable, kwargs, test.fit_kwargs, input_variable, state=state,
        validate=validate)


@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("none_first", [True, False])
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
def test_combined_running_test_(kwargs, state, validate, fit, none_first):
    test.combined_running_test(
        [test.single_target_data, test.batch_target_data], kwargs,
        test.fit_kwargs, state=state, validate=validate, fit=fit,
        none_first=none_first, aggregated=True)



from imnn.imnn.simulator_imnn import SimulatorIMNN
from imnn.imnn.aggregated_simulator_imnn import AggregatedSimulatorIMNN
from imnn.imnn.gradient_imnn import GradientIMNN
from imnn.imnn.aggregated_gradient_imnn import AggregatedGradientIMNN
from imnn.imnn.dataset_gradient_imnn import DatasetGradientIMNN
from imnn.imnn.numerical_gradient_imnn import NumericalGradientIMNN
from imnn.imnn.aggregated_numerical_gradient_imnn import \
    AggregatedNumericalGradientIMNN
from imnn.imnn.dataset_numerical_gradient_imnn import \
    DatasetNumericalGradientIMNN
from imnn.imnn.imnn import IMNN

__author__ = "Tom Charnock"
__version__ = "0.3.2"

__all__ = [
    "IMNN",
    "SimulatorIMNN",
    "AggregatedSimulatorIMNN",
    "GradientIMNN",
    "AggregatedGradientIMNN",
    "DatasetGradientIMNN",
    "NumericalGradientIMNN",
    "AggregatedNumericalGradientIMNN",
    "DatasetNumericalGradientIMNN"
]



import jax
import jax.numpy as np
from functools import partial
from imnn.imnn._aggregated_imnn import _AggregatedIMNN
from imnn.imnn import SimulatorIMNN
from imnn.utils import jacrev, value_and_jacrev, add_nested_pytrees
from imnn.utils.utils import _check_splitting


class AggregatedSimulatorIMNN(_AggregatedIMNN, SimulatorIMNN):
    """Information maximising neural network fit with simulations on-the-fly

    Defines the function to get simulations and compress them using an XLA
    compilable simulator.

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    random number generators are generated and used to generate a set of
    :math:`n_s` simulations,
    :math:`{\\bf d}^i={\\rm simulator}({\\rm seed}^i, \\theta^\\rm{fid})` at
    the fiducial model parameters, :math:`\\theta^\\rm{fid}`, and these are
    passed direrectly through a network :math:`f_{{\\bf w}}({\\bf d})` with
    network parameters :math:`{\\bf w}` to obtain network outputs
    :math:`{\\bf x}^i` and autodifferentiation is used to get the derivative of
    :math:`n_d` of these outputs with respect to the physical model parameters,
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha`, where
    :math:`\\alpha` labels the physical parameter. With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    To enable the use of large data (or networks) the whole procedure is
    aggregated. This means that the generation and passing of the simulations
    through the network is farmed out to the desired XLA devices, and
    recollected, ``n_per_device`` inputs at a time. These are then used to
    calculate the automatic gradient of the loss function with respect to the
    calculated summaries and derivatives,
    :math:`\\partial\\Lambda/\\partial{\\bf x}^i` (which is a fairly small
    computation as long as ``n_summaries`` and ``n_s`` {and ``n_d``} are not
    huge). Once this is calculated, the simulations are passed through the
    network AGAIN this time calculating the Jacobian of the network output with
    respect to the network parameters
    :math:`\\partial{\\bf x}^i/\\partial{\\bf w}` which is then combined via
    the chain rule to get

    .. math::
        \\frac{\\partial\\Lambda}{\\partial{\\bf w}} =
        \\frac{\\partial\\Lambda}{\\partial{\\bf x}^i}
        \\frac{\\partial{\\bf x}^i}{\\partial{\\bf w}}

    This can then be passed to the optimiser.

    Parameters
    ----------
    n_remaining: int
        The number simulations where only the fiducial simulations are
        calculated. This is zero if ``n_s`` is equal to ``n_d``.
    n_iterations : int
        Number of iterations through the main summarising loop
    n_remaining_iterations : int
        Number of iterations through the remaining simulations used for quick
        loops with no derivatives
    batch_shape: tuple
        The shape which ``n_d`` should be reshaped to for aggregating.
        ``n_d // (n_devices * n_per_device), n_devices, n_per_device``
    remaining_batch_shape: tuple
        The shape which ``n_s - n_d`` should be reshaped to for aggregating.
        ``(n_s - n_d) // (n_devices * n_per_device), n_devices, n_per_device``

    Methods
    -------
    simulator:
        A function for generating a simulation on-the-fly (XLA compilable)

    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, simulator, host, devices,
                 n_per_device):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also fills the
        simulation attributes (and validation if available).

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        simulator : fn
            A function that generates a single simulation from a random number
            generator and a tuple (or array) of parameter values at which to
            generate the simulations. For the purposes of use in LFI/ABC
            afterwards it is also useful for the simulator to be able to
            broadcast to a batch of simulations on the zeroth axis
            ``fn(int(2,), float([None], n_params)) ->
            float([None], input_shape)``
        host: jax.device
            The main device where the Fisher calculation is performed
        devices: list
            A list of the available jax devices (from ``jax.devices()``)
        n_per_device: int
            Number of simulations to handle at once, this should be as large as
            possible without letting the memory overflow for the best
            performance
        """
        SimulatorIMNN.__init__(
            self=self,
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser,
            simulator=simulator)
        _AggregatedIMNN.__init__(
            self=self,
            host=host,
            devices=devices,
            n_per_device=n_per_device)

    def _set_shapes(self):
        """Calculates the shapes for batching over different devices

        Raises
        ------
        ValueError
            If the difference between n_s and n_d won't scale over xla devices
        """
        self.n_remaining = self.n_s - self.n_d
        _check_splitting(self.n_remaining, "n_s - n_d", self.n_devices,
                         self.n_per_device)
        self.n_iterations = self.n_d // (self.n_devices * self.n_per_device)
        self.n_remaining_iterations = self.n_remaining \
            // (self.n_devices * self.n_per_device)
        self.batch_shape = (
            self.n_devices,
            self.n_iterations,
            self.n_per_device)
        self.remaining_batch_shape = (
            self.n_devices,
            self.n_remaining_iterations,
            self.n_per_device)

    def get_summary(self, input, w, θ, derivative=False, gradient=False):
        """ Returns a single summary of a simulation or its gradient

        Parameters
        ----------
        input : float(input_shape) or tuple
            A random number generator for a single simulation to pass through
            the network or a tuple of either (if ``gradient`` and
            ``not derivative``)

                - **dΛ_dx** *float(input_shape, n_params)* -- the derivative of
                  the loss function with respect to a network summary
                - **key** *int(2,)* -- A random number generator for a single
                  simulation

            or (if ``gradient`` and ``derivative``)

                - tuple (gradients)
                    - **dΛ_dx** *float(input_shape, n_params)* -- the
                      derivative of the loss function with respect to a network
                      summary
                    - **d2Λ_dxdθ** *float(input_shape, n_params)* -- the
                      derivative of the loss function with respect to the
                      derivative of a network summary with respect to model
                      parameters

                - **key** *int(2,)* -- A random number generator for a single
                  simulation
        w : list
            The network parameters
        θ : float(n_params,)
            The value of the parameters to generate the simulation at
            (fiducial), unused if not simulating on the fly
        derivative : bool, default=False
            Whether a derivative of the simulation with respect to model
            parameters is also passed.
        gradient : bool, default=False
            Whether to calculate the gradient with respect to model parameters

        Returns
        -------
        tuple (if ``gradient``):
            The gradient of the loss function with respect to model parameters
        float(n_summaries,) (if ``not gradient``):
            The output of the network
        float(n_summaries, n_params) (if ``derivative`` and ``not gradient``):
            The derivative of the output of the network wrt model parameters
        """
        def fn(key, w, θ):
            """ Returns a compressed simulation from the network

            Parameters
            ----------
            key : int(2,)
                The random number generator for the simulation
            w : list
                The network parameters
            θ : float(n_params,)
                The model parameters to generate the simulation at

            Returns
            ------
            float(n_summaries,):
                The summary from the output of the network
            """
            return self.model(w, self.simulator(key, θ))

        def grad_fn(key, w, θ):
            """ Derivative of the network outputs wrt to inputs and parameters

            Parameters
            ----------
            key : int(2,)
                The random number generator for the simulation
            w : list
                The network parameters
            θ : float(n_params,)
                The model parameters to generate the simulation at

            Returns
            ------
            float(n_summaries, n_params):
                The derivative of the network output with respect to the model
                parameters
            list:
                The derivative of the network outputs with respect to the
                network parameters
            """
            dx_dw, dx_dθ = jacrev(fn, argnums=(1, 2))(key, w, θ)
            return dx_dθ, dx_dw

        if derivative and gradient:
            dΛ_dx, key = input
            dΛ_dx, d2Λ_dxdθ = dΛ_dx
            d2x_dwdθ, dx_dw = jacrev(
                grad_fn, argnums=1, has_aux=True)(key, w, θ)
            return add_nested_pytrees(
                self._construct_gradient(dx_dw, aux=dΛ_dx, func="einsum"),
                self._construct_gradient(
                    d2x_dwdθ, aux=d2Λ_dxdθ, func="derivative_einsum"))
        elif derivative:
            return value_and_jacrev(fn, argnums=2)(input, w, θ)
        elif gradient:
            dΛ_dx, key = input
            dx_dw = jacrev(fn, argnums=1)(key, w, θ)
            return self._construct_gradient(dx_dw, aux=dΛ_dx, func="einsum")
        else:
            return fn(input, w, θ)

    def _collect_input(self, key, validate=False):
        """Returns the keys for generating simulations on-the-fly

        Parameters
        ----------
        key : None or int(2,)
            Random number generators for generating simulations
        validate : bool, default=False
            Whether to return the set for validation or for fitting
            (always False)
        """
        keys = np.array(jax.random.split(key, num=self.n_s))
        main = [_key for _key in keys[:self.n_d].reshape(
            self.batch_shape + (2,))]
        remaining = [_key for _key in keys[self.n_d:].reshape(
            self.remaining_batch_shape + (2,))]
        return main, remaining

    def get_summaries(self, w, key, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        Loops through the generated random number generators on each XLA device
        to generate a simulation and then pass them through the network to get
        the network outputs. These are then pushed back to the host for the
        computation of the loss function.

        The fiducial simulations which have a derivative with respect to the
        model parameters counterpart are processed first and then the remaining
        fiducial simulations are processed and concatenated.

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,)
            A random number generator for generating simulations on-the-fly
        validate : bool, default=False
            Whether to get summaries of the validation set (always False)

        Returns
        -------
        float(n_s, n_summaries):
            The set of all network outputs used to calculate the covariance
        float(n_d, n_summaries, n_params):
            The set of all the derivatives of the network outputs with respect
            to model parameters
        """
        main, remaining = self._collect_input(key, validate=validate)

        x = [np.zeros((self.n_s // (self.n_per_device * self.n_devices),
                       self.n_per_device,
                       self.n_summaries))
             for i in range(self.n_devices)]
        dx_dθ = [np.zeros((self.n_iterations,
                          self.n_per_device,
                          self.n_summaries,
                          self.n_params))
                 for i in range(self.n_devices)]

        for i in range(self.n_iterations):
            for j, (fn, dataset) in enumerate(zip(
                    self.batch_summaries_with_derivatives,
                    main)):
                _x, _dx_dθ = fn(dataset[i], w)
                x[j] = jax.ops.index_update(x[j], jax.ops.index[i], _x)
                dx_dθ[j] = jax.ops.index_update(
                    dx_dθ[j], jax.ops.index[i], _dx_dθ)

        for i in range(self.n_remaining_iterations):
            for j, (fn, dataset) in enumerate(zip(
                    self.batch_summaries, remaining)):
                x[j] = jax.ops.index_update(
                    x[j],
                    jax.ops.index[i + self.n_iterations],
                    fn(dataset[i], w))

        if self.n_devices > 1:
            x = [jax.device_put(_x, device=self.host)
                 for _x in x]
            dx_dθ = [jax.device_put(_dx_dθ, device=self.host)
                     for _dx_dθ in dx_dθ]

        x = np.stack(x, 0).reshape((self.n_s, self.n_summaries))
        dx_dθ = np.stack(dx_dθ, 0).reshape(
            (self.n_d, self.n_summaries, self.n_params))
        return x, dx_dθ

    def _split_dΛ_dx(self, dΛ_dx):
        """ Returns the gradient of loss function wrt summaries (derivatives)

        The gradient of loss function with respect to network outputs and
        their derivatives with respect to model parameters has to be reshaped
        and aggregated onto each XLA device matching the format keys are
        generated for generating simulations.

        Parameters
        ----------
        dΛ_dx : tuple
            - **dΛ_dx** *float(n_s, n_params, n_summaries)* -- the derivative
              of the loss function with respect to network summaries
            - **d2Λ_dxdθ** *float(n_d, n_summaries, n_params)* -- the
              derivative of the loss function with respect to the derivative of
              network summaries with respect to model parameters

        Return
        -----

        """
        main_gradients = [
            jax.device_put(val, device=device) for val, device in zip(
                dΛ_dx[0][:self.n_d].reshape(
                    self.batch_shape + (self.n_summaries,)),
                self.devices)]
        main_derivative_gradients = [
            jax.device_put(val, device=device) for val, device in zip(
                dΛ_dx[1].reshape(
                    self.batch_shape + (self.n_summaries, self.n_params)),
                self.devices)]
        remaining_gradients = [
            jax.device_put(val, device=device) for val, device in zip(
                dΛ_dx[0][self.n_d:].reshape(
                    self.remaining_batch_shape + (self.n_summaries,)),
                self.devices)]
        return main_gradients, main_derivative_gradients, remaining_gradients

    def get_gradient(self, dΛ_dx, w, key=None):
        """Aggregates gradients together to update the network parameters

        To avoid having to calculate the gradient with respect to all the
        simulations at once we aggregate by addition the gradient calculation
        by looping over the keys for generating the simulations again which we
        then combine with the derivative of the loss function with respect to
        the network outputs (and their derivatives with respect to the model
        parameters). Whilst this is expensive, it is necessary since we cannot
        make a stochastic estimate of the Fisher information accurately and
        therefore we need to use all the simulations available - which is
        probably too large to fit in memory.

        Parameters
        ----------
        dΛ_dx : tuple
            - **dΛ_dx** *float(n_s, n_params, n_summaries)* -- the derivative
              of the loss function with respect to network summaries
            - **d2Λ_dxdθ** *float(n_d, n_summaries, n_params)* -- the
              derivative of the loss function with respect to the derivative of
              network summaries with respect to model parameters
        w : list
            Network parameters
        key : None or int(2,)
            Random number generator

        Returns
        -------
        list:
            The gradient of the loss function with respect to the network
            parameters calculated by aggregating
        """
        main_gradients, main_derivative_gradients, remaining_gradients = \
            self._split_dΛ_dx(dΛ_dx)
        main, remaining = self._collect_input(key, validate=False)

        gradient = [
            jax.jit(
                partial(self._construct_gradient, func="zeros"),
                device=device)(w)
            for device in self.devices]

        for i in range(self.n_iterations):
            for j, (fn, dataset, grad, der_grad, device) in enumerate(
                    zip(self.batch_gradients_with_derivatives, main,
                        main_gradients, main_derivative_gradients,
                        self.devices)):
                gradient[j] = jax.jit(add_nested_pytrees, device=device)(
                    gradient[j],
                    fn(((grad[i], der_grad[i]), dataset[i]), w))
        for i in range(self.n_remaining_iterations):
            for j, (fn, dataset, grad, device) in enumerate(
                    zip(self.batch_gradients, remaining, remaining_gradients,
                        self.devices)):
                gradient[j] = jax.jit(add_nested_pytrees, device=device)(
                    gradient[j],
                    fn((grad[i], dataset[i]), w))

        return list(jax.jit(add_nested_pytrees, device=self.host)(*gradient))



from imnn.lfi.lfi import LikelihoodFreeInference
from imnn.lfi.gaussian_approximation import GaussianApproximation
from imnn.lfi.approximate_bayesian_computation import \
    ApproximateBayesianComputation
from imnn.lfi.population_monte_carlo import PopulationMonteCarlo

__author__ = "Tom Charnock"
__version__ = "0.3.2"

__all__ = [
    "LikelihoodFreeInference",
    "GaussianApproximation",
    "ApproximateBayesianComputation",
    "PopulationMonteCarlo"
]



import pytest
import jax.numpy as np
from test.numerical_gradient_defaults import numericalGradientTests
from imnn.imnn import NumericalGradientIMNN

test = numericalGradientTests(
    imnn=NumericalGradientIMNN, filename="numerical_gradient")


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable",
    [None, list(), 1., 1, np.zeros((1,)), test.rng, tuple(), (0, 0),
     (test.model[0], 0), test.bad_model, test.state])
@pytest.mark.parametrize("variable", test.kwargs.keys())
def test_initialisation_parameters_(
        variable, kwargs, input_variable, state, validate):
    test.initialise_parameters(
        variable, kwargs, input_variable, state=state, validate=validate)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable", [None, list(), 1., 1, np.zeros((1,)), test.rng])
@pytest.mark.parametrize("variable", test.fit_kwargs.keys())
def test_fit_parameters_(variable, kwargs, input_variable, state, validate):
    test.fit_parameters(
        variable, kwargs, test.fit_kwargs, input_variable, state=state,
        validate=validate)


@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("none_first", [True, False])
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
def test_combined_running_test_(kwargs, state, validate, fit, none_first):
    test.combined_running_test(
        [test.single_target_data, test.batch_target_data], kwargs,
        test.fit_kwargs, state=state, validate=validate, fit=fit,
        none_first=none_first)



import math
import jax
import jax.numpy as np
import matplotlib.pyplot as plt
from functools import partial
from imnn.utils.utils import _check_boolean, _check_type, _check_input, \
    _check_model, _check_model_output, _check_optimiser, _check_state, \
    _check_statistics_set
from imnn.utils import progress_bar


class _IMNN:
    """Information maximising neural network parent class

    This class defines the general fitting framework for information maximising
    neural networks. It includes the generic calculations of the Fisher
    information matrix from the outputs of a neural network as well as an XLA
    compilable fitting routine (with and without a progress bar). This class
    also provides a plotting routine for fitting history and a function to
    calculate the score compression of network outputs to quasi-maximum
    likelihood estimates of model parameter values.

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    simulations and :math:`n_d` derivatives with respect to physical model
    parameters are used to calculate network outputs and their derivatives
    with respect to the physical model parameters, :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha`, where
    :math:`\\alpha` labels the physical parameter. The exact details of how
    these are calculated depend on the type of available data (see list of
    different IMNN below). With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    Once the loss function is calculated the automatic gradient is then
    calculated and used to update the network parameters via the optimiser
    function. Note for large input data-sizes, large :math:`n_s` or massive
    networks the gradients may need manually accumulating via the
    :func:`~imnn.imnn._aggregated_imnn._AggregatedIMNN`.

    ``_IMNN`` is designed as the parent class for a range of specific case
    IMNNs. There is a helper function (IMNN) which should return the correct
    case when provided with the correct data. These different subclasses are:

    :func:`~imnn.SimulatorIMNN`:

        Fit an IMNN using simulations generated on-the-fly from a jax (XLA
        compilable) simulator

    :func:`~imnn.GradientIMNN`:

        Fit an IMNN using a precalculated set of fiducial simulations and their
        derivatives with respect to model parameters

    :func:`~imnn.NumericalGradientIMNN`:

        Fit an IMNN using a precalculated set of fiducial simulations and
        simulations generated using parameter values just above and below the
        fiducial parameter values to make a numerical estimate of the
        derivatives of the network outputs. Best stability is achieved when
        seeds of the simulations are matched between all parameter directions
        for the numerical derivative

    :func:`~imnn.AggregatedSimulatorIMNN`:

        ``SimulatorIMNN`` distributed over multiple jax devices and gradients
        aggregated manually. This might be necessary for very large input sizes
        as batching cannot be done when calculating the Fisher information
        matrix

    :func:`~imnn.AggregatedGradientIMNN`:

        ``GradientIMNN`` distributed over multiple jax devices and gradients
        aggregated manually. This might be necessary for very large input sizes
        as batching cannot be done when calculating the Fisher information
        matrix

    :func:`~imnn.AggregatedNumericalGradientIMNN`:

        ``NumericalGradientIMNN`` distributed over multiple jax devices and
        gradients aggregated manually. This might be necessary for very large
        input sizes as batching cannot be done when calculating the Fisher
        information matrix

    :func:`~imnn.DatasetGradientIMNN`:

        ``AggregatedGradientIMNN`` with prebuilt TensorFlow datasets

    :func:`~imnn.DatasetNumericalGradientIMNN`:

        ``AggregatedNumericalGradientIMNN`` with prebuilt TensorFlow datasets

    There are currently two other parent classes

    :func:`~imnn.imnn._aggregated_imnn.AggregatedIMNN`:

        This is the parent class which provides the fitting routine when the
        gradients of the network parameters are aggregated manually rather than
        automatically by jax. This is necessary if the size of an entire batch
        of simulations (and their derivatives with respect to model parameters)
        and the network parameters and their calculated gradients is too large
        to fit into memory. Note there is a significant performance loss from
        using the aggregation so it should only be used for these large data
        cases

    Parameters
    ----------
    n_s : int
        Number of simulations used to calculate network output covariance
    n_d : int
        Number of simulations used to calculate mean of network output
        derivative with respect to the model parameters
    n_params : int
        Number of model parameters
    n_summaries : int
        Number of summaries, i.e. outputs of the network
    input_shape : tuple
        The shape of a single input to the network
    θ_fid : float(n_params,)
        The value of the fiducial parameter values used to generate inputs
    validate : bool
        Whether a validation set is being used
    simulate : bool
        Whether input simulations are generated on the fly
    _run_with_pbar : bool
        Book keeping parameter noting that a progress bar is used when
        fitting (induces a performance hit). If ``run_with_pbar = True``
        and ``run_without_pbar = True`` then a jit compilation error will
        occur and so it is prevented
    _run_without_pbar : bool
        Book keeping parameter noting that a progress bar is not used when
        fitting. If ``run_with_pbar = True`` and ``run_without_pbar = True``
        then a jit compilation error will occur and so it is prevented
    F : float(n_params, n_params)
        Fisher information matrix calculated from the network outputs
    invF : float(n_params, n_params)
        Inverse Fisher information matrix calculated from the network outputs
    C : float(n_summaries, n_summaries)
        Covariance of the network outputs
    invC : float(n_summaries, n_summaries)
        Inverse covariance of the network outputs
    μ : float(n_summaries,)
        Mean of the network outputs
    dμ_dθ : float(n_summaries, n_params)
        Derivative of the mean of the network outputs with respect to model
        parameters
    state : :obj:state
        The optimiser state used for updating the network parameters and
        optimisation algorithm
    initial_w : list
        List of the network parameters values at initialisation (to restart)
    final_w : list
        List of the network parameters values at the end of fitting
    best_w : list
        List of the network parameters values which provide the maxmimum value
        of the determinant of the Fisher matrix
    w : list
        List of the network parameters values (either final or best depending
        on setting when calling fit(...))
    history : dict
        A dictionary containing the fitting history. Keys are
            - **detF** -- determinant of the Fisher information at the end of
              each iteration
            - **detC** -- determinant of the covariance of network outputs at
              the end of each iteration
            - **detinvC** -- determinant of the inverse covariance of network
              outputs at the end of each iteration
            - **Λ2** -- value of the covariance regularisation at the end of
              each iteration
            - **r** -- value of the regularisation coupling at the end of each
              iteration
            - **val_detF** -- determinant of the Fisher information of the
              validation data at the end of each iteration
            - **val_detC** -- determinant of the covariance of network outputs
              given the validation data at the end of each iteration
            - **val_detinvC** -- determinant of the inverse covariance of
              network outputs given the validation data at the end of each
              iteration
            - **val_Λ2** -- value of the covariance regularisation given the
              validation data at the end of each iteration
            - **val_r** -- value of the regularisation coupling given the
              validation data at the end of each iteration
            - **max_detF** -- maximum value of the determinant of the Fisher
              information on the validation data (if available)

    Methods
    -------
    model:
        Neural network as a function of network parameters and inputs
    _get_parameters:
        Function which extracts the network parameters from the state
    _model_initialiser:
        Function to initialise neural network weights from RNG and shape tuple
    _opt_initialiser:
        Function which generates the optimiser state from network parameters
    _update:
        Function which updates the state from a gradient

    Todo
    ----
    - Finish all docstrings and documentation
    - Update `NoiseNumericalGradientIMNN` to inherit from `_AggregatedIMNN`

    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state):
        """Constructor method

        Initialises all _IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float([None], input_shape)) -> float([None],
            n_summaries)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        """
        self._initialise_parameters(
            n_s, n_d, n_params, n_summaries, input_shape, θ_fid)
        self._initialise_model(model, optimiser, key_or_state)
        self._initialise_history()

    def _initialise_parameters(self, n_s, n_d, n_params, n_summaries,
                               input_shape, θ_fid):
        """Performs type checking and initialisation of class attributes

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs

        Raises
        ------
        TypeError
            Any of the parameters are not correct type
        ValueError
            Any of the parameters are ``None``
            ``Θ_fid`` has the wrong shape
        """
        self.n_s = _check_type(n_s, int, "n_s")
        self.n_d = _check_type(n_d, int, "n_d")
        self.n_params = _check_type(n_params, int, "n_params")
        self.n_summaries = _check_type(n_summaries, int, "n_summaries")
        self.input_shape = _check_type(input_shape, tuple, "input_shape")
        self.θ_fid = _check_input(θ_fid, (self.n_params,), "θ_fid")

        self.validate = False
        self.simulate = False
        self._run_with_pbar = False
        self._run_without_pbar = False

        self.F = None
        self.invF = None
        self.C = None
        self.invC = None
        self.μ = None
        self.dμ_dθ = None

        self._model_initialiser = None
        self.model = None
        self._opt_initialiser = None
        self._update = None
        self._get_parameters = None
        self.state = None
        self.initial_w = None
        self.final_w = None
        self.best_w = None
        self.w = None

        self.history = None

    def _initialise_model(self, model, optimiser, key_or_state):
        """Initialises neural network parameters or loads optimiser state

        Parameters
        ----------
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and
            the neural network as a function of network parameters and inputs
            ``fn(w: list, d: float([None], input_shape)) -> float([None],
            n_summaries)``. (Essentibly stax-like, see `jax.experimental.stax
            <https://jax.readthedocs.io/en/stable/jax.experimental.stax.html>`_
            ))
        optimiser : tuple or obj, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser

        Notes
        -----
        The design of the model follows `jax's stax module <https://jax.readth
        edocs.io/en/latest/jax.experimental.stax.html>`_ in that the model is
        encapsulated by two functions, one to initialise the network and one to
        call the model, i.e.::

            import jax
            from jax.experimental import stax

            rng = jax.random.PRNGKey(0)

            data_key, model_key = jax.random.split(rng)

            input_shape = (10,)
            inputs = jax.random.normal(data_key, shape=input_shape)

            model = stax.serial(
                stax.Dense(10),
                stax.LeakyRelu,
                stax.Dense(10),
                stax.LeakyRelu,
                stax.Dense(2))

            output_shape, initial_params = model[0](model_key, input_shape)

            outputs = model[1](initial_params, inputs)

        Note that the model used in the IMNN is assumed to be totally
        broadcastable, i.e. any batch shape can be used for inputs. This might
        require having a layer which reshapes all batch dimensions into a
        single dimension and then unwraps it at the last layer. A model such as
        that above is already fully broadcastable.

        The optimiser should follow `jax's experimental optimiser module <http
        s://jax.readthedocs.io/en/stable/jax.experimental.optimizers.html>`_ in
        that the optimiser is encapsulated by three functions, one to
        initialise the state, one to update the state from a list of gradients
        and one to extract the network parameters from the state, .i.e

        .. code-block:: python

            from jax.experimental import optimizers
            import jax.numpy as np

            optimiser = optimizers.adam(step_size=1e-3)

            initial_state = optimiser[0](initial_params)
            params = optimiser[2](initial_state)

            def scalar_output(params, inputs):
                return np.sum(model[1](params, inputs))

            counter = 0
            grad = jax.grad(scalar_output, argnums=0)(params, inputs)
            state = optimiser[1](counter, grad, state)

        This function either initialises the neural network or the state if
        passed a stateless random number generator in ``key_or_state`` or loads
        a predefined state if the state is passed to ``key_or_state``. The
        functions get mapped to the class functions

        .. code-block:: python

            self.model = model[1]
            self._model_initialiser = model[0]

            self._opt_initialiser = optimiser[0]
            self._update = optimiser[1]
            self._get_parameters = optimiser[2]

        The state is made into the ``state`` class attribute and the parameters
        are assigned to ``initial_w``, ``final_w``, ``best_w`` and ``w`` class
        attributes (where ``w`` stands for weights).

        There is some type checking done, but for freedom of choice of model
        there will be very few raised warnings.

        Raises
        ------
        TypeError
            If the random number generator is not correct, or if there is no
            possible way to construct a model or an optimiser from the passed
            parameters
        ValueError
            If any input is ``None`` or if the functions for the model or
            optimiser do not conform to the necessary specifications
        """
        self._model_initialiser, self.model = _check_model(model)
        self._opt_initialiser, self._update, self._get_parameters = \
            _check_optimiser(optimiser)
        state, key = _check_state(key_or_state)
        if key is not None:
            key = _check_input(key, (2,), "key_or_state")
            output, initial_w = self._model_initialiser(key, self.input_shape)
            _check_model_output(output, (self.n_summaries,))
            self.state = self._opt_initialiser(initial_w)
        else:
            self.state = state
            try:
                self._get_parameters(self.state)
            except Exception:
                raise TypeError("`state` is not valid for extracting " +
                                "parameters from")

        self.initial_w = self._get_parameters(self.state)
        self.final_w = self._get_parameters(self.state)
        self.best_w = self._get_parameters(self.state)
        self.w = self._get_parameters(self.state)

    def _initialise_history(self):
        """Initialises history dictionary attribute

        Notes
        -----
        The contents of the history dictionary are
            - **detF** -- determinant of the Fisher information at the end of
              each iteration
            - **detC** -- determinant of the covariance of network outputs at
              the end of each iteration
            - **detinvC** -- determinant of the inverse covariance of network
              outputs at the end of each iteration
            - **Λ2** -- value of the covariance regularisation at the end of
              each iteration
            - **r** -- value of the regularisation coupling at the end of each
              iteration
            - **val_detF** -- determinant of the Fisher information of the
              validation data at the end of each iteration
            - **val_detC** -- determinant of the covariance of network outputs
              given the validation data at the end of each iteration
            - **val_detinvC** -- determinant of the inverse covariance of
              network outputs given the validation data at the end of each
              iteration
            - **val_Λ2** -- value of the covariance regularisation given the
              validation data at the end of each iteration
            - **val_r** -- value of the regularisation coupling given the
              validation data at the end of each iteration
            - **max_detF** -- maximum value of the determinant of the Fisher
              information on the validation data (if available)

        """
        self.history = {
            "detF": np.zeros((0,)),
            "detC": np.zeros((0,)),
            "detinvC": np.zeros((0,)),
            "Λ2": np.zeros((0,)),
            "r": np.zeros((0,)),
            "val_detF": np.zeros((0,)),
            "val_detC": np.zeros((0,)),
            "val_detinvC": np.zeros((0,)),
            "val_Λ2": np.zeros((0,)),
            "val_r": np.zeros((0,)),
            "max_detF": np.float32(0.)
        }

    def _set_history(self, results):
        """Places results from fitting into the history dictionary

        Parameters
        ----------
        results : list
            List of results from fitting procedure. These are:
                - **detF** *(float(n_iterations, 2))* -- determinant of the
                  Fisher information, ``detF[:, 0]`` for training and
                  ``detF[:, 1]`` for validation
                - **detC** *(float(n_iterations, 2))* -- determinant of the
                  covariance of network outputs, ``detC[:, 0]`` for training
                  and ``detC[:, 1]`` for validation
                - **detinvC** *(float(n_iterations, 2))* -- determinant of the
                  inverse covariance of network outputs, ``detinvC[:, 0]`` for
                  training and ``detinvC[:, 1]`` for validation
                - **Λ2** *(float(n_iterations, 2))* -- value of the covariance
                  regularisation, ``Λ2[:, 0]`` for training and ``Λ2[:, 1]``
                  for validation
                - **r** *(float(n_iterations, 2))* -- value of the
                  regularisation coupling, ``r[:, 0]`` for training and
                  ``r[:, 1]`` for validation

        """
        keys = ["detF", "detC", "detinvC", "Λ2", "r"]
        for result, key in zip(results, keys):
            self.history[key] = np.hstack([self.history[key], result[:, 0]])
            if self.validate:
                self.history[f"val_{key}"] = np.hstack(
                    [self.history[f"val_{key}"], result[:, 1]])

    def _set_inputs(self, rng, max_iterations):
        """Builds list of inputs for the XLA compilable fitting routine

        Parameters
        ----------
        rng : int(2,) or None
            A stateless random number generator
        max_iterations : int
            Maximum number of iterations to run the fitting procedure for

        Notes
        -----
        The list of inputs to the routine are
            - **max_detF** *(float)* -- The maximum value of the determinant of
              the Fisher information matrix calculated so far. This is zero if
              not run before or the value from previous calls to ``fit``
            - **best_w** *(list)* -- The value of the network parameters which
              obtained the maxmimum determinant of the Fisher information
              matrix. This is the initial network parameter values if not run
              before
            - **detF** *(float(max_iterations, 1) or
              float(max_iterations, 2))* -- A container for all possible values
              of the determinant of the Fisher information matrix during each
              iteration of fitting. If there is no validation (for simulation
              on-the-fly for example) then this container has a shape of
              ``(max_iterations, 1)``, otherwise validation values are stored
              in ``detF[:, 1]``.
            - **detC** *(float(max_iterations, 1) or
              float(max_iterations, 2))* -- A container for all possible values
              of the determinant of the covariance of network outputs during
              each iteration of fitting. If there is no validation (for
              simulation on-the-fly for example) then this container has a
              shape of ``(max_iterations, 1)``, otherwise validation values are
              stored in ``detC[:, 1]``.
            - **detF** *(float(max_iterations, 1) or
              float(max_iterations, 2))* -- A container for all possible values
              of the determinant of the inverse covariance of network outputs
              during each iteration of fitting. If there is no validation (for
              simulation on-the-fly for example) then this container has a
              shape of ``(max_iterations, 1)``, otherwise validation values are
              stored in ``detinvC[:, 1]``.
            - **Λ2** *(float(max_iterations, 1) or
              float(max_iterations, 2))* -- A container for all possible values
              of the covariance regularisation during each iteration of
              fitting. If there is no validation (for simulation on-the-fly for
              example) then this container has a shape of
              ``(max_iterations, 1)``, otherwise validation values are stored
              in ``Λ2[:, 1]``.
            - **r** *(float(max_iterations, 1) or
              float(max_iterations, 2))* -- A container for all possible values
              of the regularisation coupling strength during each iteration of
              fitting. If there is no validation (for simulation on-the-fly for
              example) then this container has a shape of
              ``(max_iterations, 1),`` otherwise validation values are stored
              in ``r[:, 1]``.
            - **counter** *(int)* -- Iteration counter used to note whether the
              while loop reaches ``max_iterations``. If not, the history
              objects (above) get truncated to length ``counter``. This starts
              with value zero
            - **patience_counter** *(int)* -- Counts the number of iterations
              where there is no increase in the value of the determinant of the
              Fisher information matrix, used for early stopping. This starts
              with value zero
            - **state** *(:obj:state)* -- The current optimiser state used for
              updating the network parameters and optimisation algorithm
            - **rng** *(int(2,))* -- A stateless random number generator which
              gets updated on each iteration

        Todo
        ----
        ``rng`` is currently only used for on-the-fly simulation but could
        easily be updated to allow for stochastic models
        """
        if self.validate:
            shape = (max_iterations, 2)
        else:
            shape = (max_iterations, 1)
        return (
            self.history["max_detF"], self.best_w, np.zeros(shape),
            np.zeros(shape), np.zeros(shape), np.zeros(shape), np.zeros(shape),
            np.int32(0), np.int32(0), self.state, rng)

    def fit(self, λ, ϵ, rng=None, patience=100, min_iterations=100,
            max_iterations=int(1e5), print_rate=None, best=True):
        """Fitting routine for the IMNN

        Parameters
        ----------
        λ : float
            Coupling strength of the regularisation
        ϵ : float
            Closeness criterion describing how close to the 1 the determinant
            of the covariance (and inverse covariance) of the network outputs
            is desired to be
        rng : int(2,) or None, default=None
            Stateless random number generator
        patience : int, default=10
            Number of iterations where there is no increase in the value of the
            determinant of the Fisher information matrix, used for early
            stopping
        min_iterations : int, default=100
            Number of iterations that should be run before considering early
            stopping using the patience counter
        max_iterations : int, default=int(1e5)
            Maximum number of iterations to run the fitting procedure for
        print_rate : int or None, default=None,
            Number of iterations before updating the progress bar whilst
            fitting. There is a performance hit from updating the progress bar
            more often and there is a large performance hit from using the
            progress bar at all. (Possible ``RET_CHECK`` failure if
            ``print_rate`` is not ``None`` when using GPUs).
            For this reason it is set to None as default
        best : bool, default=True
            Whether to set the network parameter attribute ``self.w`` to the
            parameter values that obtained the maximum determinant of
            the Fisher information matrix or the parameter values at the final
            iteration of fitting

        Example
        -------

        We are going to summarise the mean and variance of some random Gaussian
        noise with 10 data points per example using a SimulatorIMNN. In this
        case we are going to generate the simulations on-the-fly with a
        simulator written in jax (from the examples directory). We will use
        1000 simulations to estimate the covariance of the network outputs and
        the derivative of the mean of the network outputs with respect to the
        model parameters (Gaussian mean and variance) and generate the
        simulations at a fiducial μ=0 and Σ=1. The network will be a stax model
        with hidden layers of ``[128, 128, 128]`` activated with leaky relu and
        outputting 2 summaries. Optimisation will be via Adam with a step size
        of ``1e-3``. Rather arbitrarily we'll set the regularisation strength
        and covariance identity constraint to λ=10 and ϵ=0.1 (these are
        relatively unimportant for such an easy model).

        .. code-block:: python

            import jax
            import jax.numpy as np
            from jax.experimental import stax, optimizers
            from imnn import SimulatorIMNN

            rng = jax.random.PRNGKey(0)

            n_s = 1000
            n_d = 1000
            n_params = 2
            n_summaries = 2
            input_shape = (10,)
            simulator_args = {"input_shape": input_shape}
            θ_fid = np.array([0., 1.])

            def simulator(rng, θ):
                return θ[0] + jax.random.normal(
                    rng, shape=input_shape) * np.sqrt(θ[1])

            model = stax.serial(
                stax.Dense(128),
                stax.LeakyRelu,
                stax.Dense(128),
                stax.LeakyRelu,
                stax.Dense(128),
                stax.LeakyRelu,
                stax.Dense(n_summaries))
            optimiser = optimizers.adam(step_size=1e-3)

            λ = 10.
            ϵ = 0.1

            model_key, fit_key = jax.random.split(rng)

            imnn = SimulatorIMNN(
                n_s=n_s, n_d=n_d, n_params=n_params, n_summaries=n_summaries,
                input_shape=input_shape, θ_fid=θ_fid, model=model,
                optimiser=optimiser, key_or_state=model_key,
                simulator=simulator)

            imnn.fit(λ, ϵ, rng=fit_key, min_iterations=1000, patience=250,
                     print_rate=None)


        Notes
        -----
        A minimum number of interations should be be run before stopping based
        on a maximum determinant of the Fisher information achieved since the
        loss function has dual objectives. Since the determinant of the
        covariance of the network outputs is forced to 1 quickly, this can be
        at the detriment to the value of the determinant of the Fisher
        information matrix early in the fitting procedure. For this reason
        starting early stopping after the covariance has converged is advised.
        This is not currently implemented but could be considered in the
        future.

        The best fit network parameter values are probably not the most
        representative set of parameters when simulating on-the-fly since there
        is a high chance of a statistically overly-informative set of data
        being generated. Instead, if using :func:`~imnn.SimulatorIMNN.fit()`
        consider using ``best=False`` which sets ``self.w=self.final_w`` which
        are the network parameter values obtained in the last iteration. Also
        consider using a larger ``patience`` value if using
        :func:`~imnn.SimulatorIMNN.fit()` to overcome the fact that a flukish
        high value for the determinant might have been obtained due to the
        realisation of the dataset.

        Due to some unusual thing, that I can't work out, there is a massive
        performance hit when calling ``jax.jit(self._fit)`` compared with
        directly decorating ``_fit`` with
        ``@partial(jax.jit(static_argnums=0))``. Unfortunately this means
        having to duplicate ``_fit`` to include a version where the loop
        condition is decorated with a progress bar because the ``tqdm``
        module cannot use a jitted tracer. If the progress bar is not used then
        the fully decorated jitted ``_fit`` function is used and it is super
        quick. Otherwise, just the body of the loop is jitted so that the
        condition function can be decorated by the progress bar (at the
        expense of a performance hit). I imagine that something can be improved
        here.

        There is a chance of a ``RET_CHECK`` failure when using the progress
        bar on GPUs (this doesn't seem to be a problem on CPUs). If this is the
        case then `print_rate=None` should be used

        Methods
        -------
        _fit:
            Main fitting function implemented as a ``jax.lax.while_loop``
        _fit_pbar:
            Main fitting function as a ``jax.lax.while_loop`` with progress bar

        Raises
        ------
        TypeError
            If any input has the wrong type
        ValueError
            If any input (except ``rng`` and ``print_rate``) are ``None``
        ValueError
            If ``rng`` has the wrong shape
        ValueError
            If ``rng`` is ``None`` but simulating on-the-fly
        ValueError
            If calling fit with ``print_rate=None`` after previous call with
            ``print_rate`` as an integer value
        ValueError
            If calling fit with ``print_rate`` as an integer after previous
            call with ``print_rate=None``

        Todo
        ----
        - ``rng`` is currently only used for on-the-fly simulation but could
          easily be updated to allow for stochastic models
        - Automatic detection of convergence based on value ``r`` when early
          stopping can be started
        """
        @jax.jit
        def _fit(inputs):
            """Main fitting function implemented as a jax.lax.while_loop

            This is the ``jax.jit`` decorated fitting routine implemented all
            within a ``jax.lax.while_loop``. The details of the stopping
            condition and the fitting procedure are in :func:~`_fit_cond`
            and :func:~`_fit`.

            Notes
            -----
            This function is almost exactly duplicated in :func:~`_fit_pbar`
            where the ``jax.jit`` decoration is only on :func:~`self._fit`
            instead of the whole function and :func:~`self._fit_cond` is
            decorated with a progress bar. Note that there is a massive
            performance hit from using the progress bar. Also, there seems to
            be a significant performance hit from using ``jax.jit(self._fit)``
            rather than decorating it - I'm not sure why this is.

            Parameters
            ----------
            inputs : tuple
                loop variables:
                - **max_detF** *(float)* -- Maximum value of the determinant of
                  the Fisher information matrix calculated so far
                - **best_w** *(list)* -- Value of the network parameters which
                  obtained the maxmimum determinant of the Fisher information
                  matrix
                - **detF** *(float(max_iterations, 1)
                  or float(max_iterations, 2))* -- History of the determinant
                  of the Fisher information matrix
                - **detC** *(float(max_iterations, 1)
                  or float(max_iterations, 2))* -- History of the determinant
                  of the covariance of network outputs
                - **detinvC** *(float(max_iterations, 1)
                  or float(max_iterations, 2))* -- History of the determinant
                  of the inverse covariance of network outputs
                - **Λ2** *(float(max_iterations, 1)
                  or float(max_iterations, 2))* -- History of the covariance
                  regularisation
                - **r** *(float(max_iterations, 1)
                  or float(max_iterations, 2))* -- History of the
                  regularisation coupling strength
                - **counter** *(int)* -- While loop iteration counter
                - **patience_counter** *(int)* -- Number of iterations where
                  there is no increase in the value of the determinant of the
                  Fisher information matrix
                - **state** *(:obj: state)* -- Optimiser state used for
                  updating the network parameters and optimisation algorithm
                - **rng** *(int(2,))* -- Stateless random number generator

            Returns
            -------
            tuple:
                loop variables (described in Parameters)
            """
            return jax.lax.while_loop(
                partial(self._fit_cond, patience=patience,
                        max_iterations=max_iterations),
                partial(self._fit, λ=λ, α=α, min_iterations=min_iterations),
                inputs)

        def _fit_pbar(inputs):
            """Main fitting function as a jax.lax.while_loop with progress bar

            This is the ``jax.jit`` decorated fitting routine implemented all
            within a ``jax.lax.while_loop``. The details of the stopping
            condition and the fitting procedure are in :func:~`_fit_cond`
            and :func:~`_fit`.

            Notes
            -----
            This function is almost exactly duplicated in :func:~`_fit` but
            here the ``jax.jit`` decoration is only on :func:~`self._fit`
            instead of the whole function and :func:~`self._fit_cond` is
            decorated with a progress bar. Note that there is a massive
            performance hit from using the progress bar. Also, there seems to
            be a significant performance hit from using ``jax.jit(self._fit)``
            rather than decorating it - I'm not sure why this is.
            """
            return jax.lax.while_loop(
                progress_bar(max_iterations, patience, print_rate)(
                    partial(self._fit_cond, patience=patience,
                            max_iterations=max_iterations)),
                jax.jit(
                    partial(self._fit, λ=λ, α=α,
                            min_iterations=min_iterations)),
                inputs)

        λ = _check_type(λ, float, "λ")
        ϵ = _check_type(ϵ, float, "ϵ")
        α = self.get_α(λ, ϵ)
        patience = _check_type(patience, int, "patience")
        min_iterations = _check_type(min_iterations, int, "min_iterations")
        max_iterations = _check_type(max_iterations, int, "max_iterations")
        best = _check_boolean(best, "best")
        if self.simulate and (rng is None):
            raise ValueError("`rng` is necessary when simulating.")
        rng = _check_input(rng, (2,), "rng", allow_None=True)
        inputs = self._set_inputs(rng, max_iterations)
        if print_rate is None:
            if self._run_with_pbar:
                raise ValueError(
                    "Cannot run IMNN without progress bar after running " +
                    "with progress bar. Either set `print_rate` to an int " +
                    "or reinitialise the IMNN.")
            else:
                self._run_without_pbar = True
                results = _fit(inputs)
        else:
            if self._run_without_pbar:
                raise ValueError(
                    "Cannot run IMNN with progress bar after running " +
                    "without progress bar. Either set `print_rate` to None " +
                    "or reinitialise the IMNN.")
            else:
                print_rate = _check_type(print_rate, int, "print_rate")
                self._run_with_pbar = True
                results = _fit_pbar(inputs)
        self.history["max_detF"] = results[0]
        self.best_w = results[1]
        self._set_history(
            (results[2][:results[7]],
             results[3][:results[7]],
             results[4][:results[7]],
             results[5][:results[7]],
             results[6][:results[7]]))
        if len(results) == 11:
            self.state = results[-2]
        self.final_w = self._get_parameters(self.state)
        if best:
            w = self.best_w
        else:
            w = self.final_w
        self.set_F_statistics(w, key=rng)

    def _get_fitting_keys(self, rng):
        """Generates random numbers for simulation generation if needed

        Parameters
        ----------
        rng : int(2,) or None
            A random number generator

        Returns
        -------
        int(2,), int(2,), int(2,) or None, None, None:
            A new random number generator and random number generators for
            training and validation, or empty values
        """
        if rng is not None:
            return jax.random.split(rng, num=3)
        else:
            return None, None, None

    def get_α(self, λ, ϵ):
        """Calculate rate parameter for regularisation from closeness criterion

        Parameters
        ----------
        λ : float
            coupling strength of the regularisation
        ϵ : float
            closeness criterion describing how close to the 1 the determinant
            of the covariance (and inverse covariance) of the network outputs
            is desired to be

        Returns
        -------
        float:
            The steepness of the tanh-like function (or rate) which determines
            how fast the determinant of the covariance of the network outputs
            should be pushed to 1
        """
        return - math.log(ϵ * (λ - 1.) + ϵ ** 2. / (1 + ϵ)) / ϵ

    def _fit(self, inputs, λ=None, α=None, min_iterations=None):
        """Single iteration fitting algorithm

        This function performs the network parameter updates first getting
        any necessary random number generators for simulators and then
        extracting the network parameters from the state. These parameters
        are used to calculate the gradient with respect to the network
        parameters of the loss function (see _IMNN class docstrings).
        Once the loss function is calculated the gradient is then used to
        update the network parameters via the optimiser function and the
        current iterations statistics are saved to the history arrays. If
        validation is used (recommended for ``GradientIMNN`` and
        ``NumericalGradientIMNN``) then all necessary statistics to
        calculate the loss function are calculated and pushed to the
        history arrays.

        The ``patience_counter`` is increased if the value of determinant
        of the Fisher information matrix does not increase over the
        previous iterations upto ``patience`` number of iterations at which
        point early stopping occurs, but only if the number of iterations
        so far performed is greater than a specified ``min_iterations``.

        Parameters
        ----------
        inputs : tuple
            - **max_detF** *(float)* -- Maximum value of the determinant of the
              Fisher information matrix calculated so far
            - **best_w** *(list)* -- Value of the network parameters which
              obtained the maxmimum determinant of the Fisher information
              matrix
            - **detF** *(float(max_iterations, 1)
              or float(max_iterations, 2))* -- History of the determinant of
              the Fisher information matrix
            - **detC** *(float(max_iterations, 1) or float(max_iterations, 2))*
              -- History of the determinant of the covariance of network
              outputs
            - **detinvC** *(float(max_iterations, 1)
              or float(max_iterations, 2))* -- History of the determinant of
              the inverse covariance of network outputs
            - **Λ2** *(float(max_iterations, 1) or float(max_iterations, 2))*
              -- History of the covariance regularisation
            - **r** *(float(max_iterations, 1) or float(max_iterations, 2))*
              -- History of the regularisation coupling strength
            - **counter** *(int)* -- While loop iteration counter
            - **patience_counter** *(int)* -- Number of iterations where there
              is no increase in the value of the determinant of the Fisher
              information matrix
            - **state** *(:obj: state)* -- Optimiser state used for updating
              the network parameters and optimisation algorithm
            - **rng** *(int(2,))* -- Stateless random number generator

        λ : float
            Coupling strength of the regularisation
        α : float
            Rate parameter for regularisation coupling
        min_iterations : int
            Number of iterations that should be run before considering early
            stopping using the patience counter

        Returns
        -------
        tuple:
            loop variables (described in Parameters)
        """
        max_detF, best_w, detF, detC, detinvC, Λ2, r, \
            counter, patience_counter, state, rng = inputs
        rng, training_key, validation_key = self._get_fitting_keys(rng)
        w = self._get_parameters(state)
        grad, results = jax.grad(
            self._get_loss, argnums=0, has_aux=True)(w, λ, α, training_key)
        state = self._update(counter, grad, state)
        w = self._get_parameters(state)
        detF, detC, detinvC, Λ2, r = self._update_history(
            results, (detF, detC, detinvC, Λ2, r), counter, 0)
        if self.validate:
            F, C, invC, *_ = self._get_F_statistics(
                w, key=validation_key, validate=True)
            _Λ2 = self._get_regularisation(C, invC)
            _r = self._get_regularisation_strength(_Λ2, λ, α)
            results = (F, C, invC, _Λ2, _r)
            detF, detC, detinvC, Λ2, r = self._update_history(
                results, (detF, detC, detinvC, Λ2, r), counter, 1)
        _detF = np.linalg.det(results[0])
        patience_counter, counter, _, max_detF, __, best_w = \
            jax.lax.cond(
                np.greater(_detF, max_detF),
                self._update_loop_vars,
                lambda inputs: self._check_loop_vars(inputs, min_iterations),
                (patience_counter, counter, _detF, max_detF, w, best_w))
        return (max_detF, best_w, detF, detC, detinvC, Λ2, r,
                counter + np.int32(1), patience_counter, state, rng)

    def _fit_cond(self, inputs, patience, max_iterations):
        """Stopping condition for the fitting loop

        The stopping conditions due to reaching ``max_iterations`` or the
        patience counter reaching ``patience`` due to ``patience_counter``
        number of iterations without increasing the determinant of the
        Fisher information matrix.

        Parameters
        ----------
        inputs : tuple
            - **max_detF** *(float)* -- Maximum value of the determinant of the
              Fisher information matrix calculated so far
            - **best_w** *(list)* -- Value of the network parameters which
              obtained the maxmimum determinant of the Fisher information
              matrix
            - **detF** *(float(max_iterations, 1)
              or float(max_iterations, 2))* -- History of the determinant of
              the Fisher information matrix
            - **detC** *(float(max_iterations, 1) or float(max_iterations, 2))*
              -- History of the determinant of the covariance of network
              outputs
            - **detinvC** *(float(max_iterations, 1)
              or float(max_iterations, 2))* -- History of the determinant of
              the inverse covariance of network outputs
            - **Λ2** *(float(max_iterations, 1) or float(max_iterations, 2))*
              -- History of the covariance regularisation
            - **r** *(float(max_iterations, 1) or float(max_iterations, 2))*
              -- History of the regularisation coupling strength
            - **counter** *(int)* -- While loop iteration counter
            - **patience_counter** *(int)* -- Number of iterations where there
              is no increase in the value of the determinant of the Fisher
              information matrix
            - **state** *(:obj: state)* -- Optimiser state used for updating
              the network parameters and optimisation algorithm
            - **rng** *(int(2,))* -- Stateless random number generator

        patience : int
            Number of iterations to stop the fitting when there is no increase
            in the value of the determinant of the Fisher information matrix
        max_iterations : int
        Maximum number of iterations to run the fitting procedure for

        Returns
        -------
        bool:
            True if either the ``patience_counter`` has not reached the
            ``patience`` criterion or if the ``counter`` has not reached
            ``max_iterations``
        """
        return np.logical_and(
            np.less(inputs[-3], patience),
            np.less(inputs[-4], max_iterations))

    def _update_loop_vars(self, inputs):
        """Updates input parameters if ``max_detF`` is increased

        If the determinant of the Fisher information matrix calculated
        in a given iteration is larger than the ``max_detF`` calculated
        so far then the ``patience_counter`` is reset to zero and the
        ``max_detF`` is replaced with the current value of ``detF`` and
        the network parameters in this iteration replace the previous
        parameters which obtained the highest determinant of the Fisher
        information, ``best_w``.

        Parameters
        ----------
        inputs : tuple
            - **patience_counter** *(int)* -- Number of iterations where
              there is no increase in the value of the determinant of the
              Fisher information matrix
            - **counter** *(int)* -- While loop iteration counter
            - **detF** *(float(max_iterations, 1)
              or float(max_iterations, 2))* -- History of the determinant
              of the Fisher information matrix
            - **max_detF** *(float)* -- Maximum value of the determinant of
              the Fisher information matrix calculated so far
            - **w** *(list)* -- Value of the network parameters which in
              current iteration
            - **best_w** *(list)* -- Value of the network parameters which
              obtained the maxmimum determinant of the Fisher information
              matrix

        Returns
        -------
        tuple:
            (described in Parameters)
        """
        patience_counter, counter, detF, max_detF, w, best_w = inputs
        return (np.int32(0), counter, detF, detF, w, w)

    def _check_loop_vars(self, inputs, min_iterations):
        """Updates ``patience_counter`` if ``max_detF`` not increased

        If the determinant of the Fisher information matrix calculated
        in a given iteration is not larger than the ``max_detF``
        calculated so far then the ``patience_counter`` is increased by
        one as long as the number of iterations is greater than the
        minimum number of iterations that should be run.

        Parameters
        ----------
        inputs : tuple
            - **patience_counter** *(int)* -- Number of iterations where
              there is no increase in the value of the determinant of the
              Fisher information matrix
            - **counter** *(int)* -- While loop iteration counter
            - **detF** *(float(max_iterations, 1)
              or float(max_iterations, 2))* -- History of the determinant
              of the Fisher information matrix
            - **max_detF** *(float)* -- Maximum value of the determinant of
              the Fisher information matrix calculated so far
            - **w** *(list)* -- Value of the network parameters which in
              current iteration
            - **best_w** *(list)* -- Value of the network parameters which
              obtained the maxmimum determinant of the Fisher information
              matrix
        min_iterations : int
            Number of iterations that should be run before considering early
            stopping using the patience counter

        Returns
        -------
        tuple:
            (described in Parameters)
        """
        patience_counter, counter, detF, max_detF, w, best_w = inputs
        patience_counter = jax.lax.cond(
            np.greater(counter, min_iterations),
            lambda patience_counter: patience_counter + np.int32(1),
            lambda patience_counter: patience_counter,
            patience_counter)
        return (patience_counter, counter, detF, max_detF, w, best_w)

    def _update_history(self, inputs, history, counter, ind):
        """Puts current fitting statistics into history arrays

        Parameters
        ----------
        inputs : tuple
            Fitting statistics calculated on a single iteration
                - **F** *(float(n_params, n_params))* -- Fisher information
                  matrix
                - **C** *(float(n_summaries, n_summaries))* -- Covariance of
                  network outputs
                - **invC** *(float(n_summaries, n_summaries))* -- Inverse
                  covariance of network outputs
                - **_Λ2** *(float)* -- Covariance regularisation
                - **_r** *(float)* -- Regularisation coupling strength
        history : tuple
            History arrays containing fitting statistics for each iteration
                - **detF** *(float(max_iterations, 1) or
                  float(max_iterations, 2))* -- Determinant of the Fisher
                  information matrix
                - **detC** *(float(max_iterations, 1) or
                  float(max_iterations, 2))* -- Determinant of the covariance
                  of network outputs
                - **detinvC** *(float(max_iterations, 1)
                  or float(max_iterations, 2))* -- Determinant of the inverse
                  covariance of network outputs
                - **Λ2** *(float(max_iterations, 1) or
                  float(max_iterations, 2))* -- Covariance regularisation
                - **r** *(float(max_iterations, 1) or
                  float(max_iterations, 2))* -- Regularisation coupling
                  strength
        counter : int
            Current iteration to insert a single iteration statistics into the
            history
        ind : int
            Values of either 0 (fitting) or 1 (validation) to separate the
            fitting and validation historys

        Returns
        -------
        float(max_iterations, 1) or float(max_iterations, 2):
            History of the determinant of the Fisher information matrix
        float(max_iterations, 1) or float(max_iterations, 2):
            History of the determinant of the covariance of network outputs
        float(max_iterations, 1) or float(max_iterations, 2):
            History of the determinant of the inverse covariance of network
            outputs
        float(max_iterations, 1) or float(max_iterations, 2):
            History of the covariance regularisation
        float(max_iterations, 1) or float(max_iterations, 2):
            History of the regularisation coupling strength
        """
        F, C, invC, _Λ2, _r = inputs
        detF, detC, detinvC, Λ2, r = history
        detF = jax.ops.index_update(
            detF,
            jax.ops.index[counter, ind],
            np.linalg.det(F))
        detC = jax.ops.index_update(
            detC,
            jax.ops.index[counter, ind],
            np.linalg.det(C))
        detinvC = jax.ops.index_update(
            detinvC,
            jax.ops.index[counter, ind],
            np.linalg.det(invC))
        Λ2 = jax.ops.index_update(
            Λ2,
            jax.ops.index[counter, ind],
            _Λ2)
        r = jax.ops.index_update(
            r,
            jax.ops.index[counter, ind],
            _r)
        return detF, detC, detinvC, Λ2, r

    def _slogdet(self, matrix):
        """Combined summed logarithmic determinant

        Parameters
        ----------
        matrix : float(n, n)
            An n x n matrix to calculate the summed logarithmic determinant of

        Returns
        -------
        float:
            The summed logarithmic determinant multiplied by its sign
        """
        lndet = np.linalg.slogdet(matrix)
        return lndet[0] * lndet[1]

    def _construct_derivatives(self, derivatives):
        """Builds derivatives of the network outputs wrt model parameters

        An empty directive in ``_IMNN``, ``SimulatorIMNN`` and ``GradientIMNN``
        but necessary to construct correct shaped derivatives when using
        ``NumericalGradientIMNN``.

        Parameters
        ----------
        derivatives : float(n_d, n_summaries, n_params)
            The derivatives of the network ouputs with respect to the model
            parameters

        Returns
        -------
        float(n_d, n_summaries, n_params):
            The derivatives of the network ouputs with respect to the model
            parameters
        """
        return derivatives

    def set_F_statistics(self, w=None, key=None, validate=True):
        """Set necessary attributes for calculating score compressed summaries

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=True
            Whether to calculate Fisher information using the validation set
        """
        if validate and ((not self.validate) and (not self.simulate)):
            validate = False
        if w is not None:
            self.w = w
        self.F, self.C, self.invC, self.dμ_dθ, self.μ = \
            self._get_F_statistics(key=key, validate=validate)
        self.invF = np.linalg.inv(self.F)

    def _get_F_statistics(self, w=None, key=None, validate=False):
        """Calculates the Fisher information and returns all statistics used

        First gets the summaries and derivatives and then uses them to
        calculate the Fisher information matrix from the outputs and return all
        the necessary constituents to calculate the Fisher information (which)
        are needed for the score compression or the regularisation of the loss
        function.

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=True
            Whether to calculate Fisher information using the validation set

        Returns
        -------
        tuple:
            - **F** *(float(n_params, n_params))* -- Fisher information matrix
            - **C** *(float(n_summaries, n_summaries))* -- Covariance of
              network outputs
            - **invC** *(float(n_summaries, n_summaries))* -- Inverse
              covariance of network outputs
            - **dμ_dθ** *(float(n_summaries, n_params))* -- The derivative of
              the mean of network outputs with respect to model parameters
            - **μ** *(float(n_summaries,))* -- The mean of the network outputs

        """
        if w is None:
            w = self.w
        summaries, derivatives = self.get_summaries(
            w=w, key=key, validate=validate)
        return self._calculate_F_statistics(summaries, derivatives)

    @partial(jax.jit, static_argnums=0)
    def _calculate_F_statistics(self, summaries, derivatives):
        """Calculates the Fisher information matrix from network outputs

        If the numerical derivative is being calculated then the derivatives
        are first constructed. If the mean is to be returned (for use in score
        compression), this is calculated and pushed to the results tuple.
        Then the covariance of the summaries is taken and inverted and the mean
        of the derivative of network summaries with respect to the model
        parameters is found and these are used to calculate the Gaussian form
        of the Fisher information matrix.

        Parameters
        ----------
        summaries : float(n_s, n_summaries)
            The network outputs
        derivatives : float(n_d, n_summaries, n_params)
            The derivative of the network outputs wrt the model parameters.
            Note that when ``NumericalGradientIMNN`` is being used the shape is
            ``float(n_d, 2, n_params, n_summaries)`` which is then constructed
            into the the numerical derivative in ``_construct_derivatives``.

        Returns
        -------
        tuple:
            - **F** *(float(n_params, n_params))* -- Fisher information matrix
            - **C** *(float(n_summaries, n_summaries))* -- Covariance of
              network outputs
            - **invC** *(float(n_summaries, n_summaries))* -- Inverse
              covariance of network outputs
            - **dμ_dθ** *(float(n_summaries, n_params))* -- The derivative of
              the mean of network outputs with respect to model parameters
            - **μ** *(float(n_summaries))* -- The mean of the
              network outputs
        """
        derivatives = self._construct_derivatives(derivatives)
        μ = np.mean(summaries, axis=0)
        C = np.cov(summaries, rowvar=False)
        if self.n_summaries == 1:
            C = np.array([[C]])
        invC = np.linalg.inv(C)
        dμ_dθ = np.mean(derivatives, axis=0)
        F = np.einsum("ij,ik,kl->jl", dμ_dθ, invC, dμ_dθ)
        return (F, C, invC, dμ_dθ, μ)

    @partial(jax.jit, static_argnums=0)
    def _get_regularisation_strength(self, Λ2, λ, α):
        """Coupling strength of the regularisation (amplified sigmoid)

        To dynamically turn off the regularisation when the scale of the
        covariance is set to approximately the identity matrix, a smooth
        sigmoid conditional on the value of the regularisation is used. The
        rate, α, is calculated from a closeness condition of the covariance
        (and the inverse covariance) to the identity matrix using ``get_α``.

        Parameters
        ----------
        Λ2 : float
            Covariance regularisation
        λ : float
            Coupling strength of the regularisation
        α : float
            Calculate rate parameter for regularisation from ϵ criterion

        Returns
        -------
        float:
            Smooth, dynamic regularisation strength
        """
        return λ * Λ2 / (Λ2 + np.exp(-α * Λ2))

    @partial(jax.jit, static_argnums=0)
    def _get_regularisation(self, C, invC):
        """Difference of the covariance (and its inverse) from identity

        The negative logarithm of the determinant of the Fisher information
        matrix needs to be regularised to fix the scale of the network outputs
        since any linear rescaling of a sufficient statistic is also a
        sufficient statistic. We choose to fix this scale by constraining the
        covariance of network outputs as

        .. math::
            \\Lambda_2 = ||\\bf{C}-\\bf{I}|| + ||\\bf{C}^{-1}-\\bf{I}||

        One benefit of choosing this constraint is that it forces the
        covariance to be approximately parameter independent which justifies
        choosing the covariance independent Gaussian Fisher information.


        Parameters
        ----------
        C : float(n_summaries, n_summaries)
            Covariance of the network ouputs
        invC : float(n_summaries, n_summaries)
            Inverse covariance of the network ouputs

        Returns
        -------
        float:
            Regularisation loss terms for the distance of the covariance and
            its determinant from the identity matrix
        """
        return np.linalg.norm(C - np.eye(self.n_summaries)) + \
            np.linalg.norm(invC - np.eye(self.n_summaries))

    def _get_loss(self, w, λ, α, key=None):
        """Calculates the loss function and returns auxillary variables

        First gets the summaries and derivatives and then uses them to
        calculate the loss function. This function is separated to be able to
        use ``jax.grad`` directly rather than calculating the derivative of the
        summaries as is done with ``_AggregatedIMNN``.

        Parameters
        ----------
        w : list
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        λ : float
            Coupling strength of the regularisation
        α : float
            Calculate rate parameter for regularisation from ϵ criterion
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly

        Returns
        -------
        float:
            Value of the regularised loss function
        tuple:
            Fitting statistics calculated on a single iteration
                - **F** *(float(n_params, n_params))* -- Fisher information
                  matrix
                - **C** *(float(n_summaries, n_summaries))* -- Covariance of
                  network outputs
                - **invC** *(float(n_summaries, n_summaries))* -- Inverse
                  covariance of network outputs
                - **Λ2** *(float)* -- Covariance regularisation
                - **r** *(float)* -- Regularisation coupling strength

        """
        summaries, derivatives = self.get_summaries(w=w, key=key)
        return self._calculate_loss(summaries, derivatives, λ, α)

    def _calculate_loss(self, summaries, derivatives, λ, α):
        """Calculates the loss function from network summaries and derivatives

        Parameters
        ----------
        summaries : float(n_s, n_summaries)
            The network outputs
        derivatives : float(n_d, n_summaries, n_params)
            The derivative of the network outputs wrt the model parameters.
            Note that when ``NumericalGradientIMNN`` is being used the shape is
            ``float(n_d, 2, n_params, n_summaries)`` which is then constructed
            into the the numerical derivative in ``_construct_derivatives``.
        λ : float
            Coupling strength of the regularisation
        α : float
            Calculate rate parameter for regularisation from ϵ criterion

        Returns
        -------
        float:
            Value of the regularised loss function
        tuple:
            Fitting statistics calculated on a single iteration
                - **F** *(float(n_params, n_params))* -- Fisher information
                  matrix
                - **C** *(float(n_summaries, n_summaries))* -- Covariance of
                  network outputs
                - **invC** *(float(n_summaries, n_summaries))* -- Inverse
                  covariance of network outputs
                - **Λ2** *(float)* -- Covariance regularisation
                - **r** *(float)* -- Regularisation coupling strength

        """
        F, C, invC, dμ_dθ, _ = self._calculate_F_statistics(
            summaries, derivatives)
        lndetF = self._slogdet(F)
        Λ2 = self._get_regularisation(C, invC)
        r = self._get_regularisation_strength(Λ2, λ, α)
        return - lndetF + r * Λ2, (F, C, invC, Λ2, r)

    def get_summaries(self, w=None, key=None, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=False
            Whether to get summaries of the validation set

        Raises
        ------
        ValueError
        """
        raise ValueError("`get_summaries` not implemented")

    def get_estimate(self, d):
        """Calculate score compressed parameter estimates from network outputs

        Using score compression we can get parameter estimates under the
        transformation

        .. math::
            \\hat{\\boldsymbol{\\theta}}_\\alpha=\\theta^{\\rm{fid}}_\\alpha+
            \\bf{F}^{-1}_{\\alpha\\beta}\\frac{\\partial\\mu_i}{\\partial
            \\theta_\\beta}\\bf{C}^{-1}_{ij}(x(\\bf{w}, \\bf{d})-\\mu)_j

        where :math:`x_j` is the :math:`j` output of the network with network
        parameters :math:`\\bf{w}` and input data :math:`\\bf{d}`.

        Examples
        --------
        Assuming that an IMNN has been fit (as in the example in
        :py:meth:`imnn.imnn._imnn.IMNN.fit`) then we can obtain a
        pseudo-maximum likelihood estimate of some target data (which is
        generated with parameter values μ=1, Σ=2) using

        .. code-block:: python

            rng, target_key = jax.random.split(rng)
            target_data = model_simulator(target_key, np.array([1., 2.]))

            imnn.get_estimate(target_data)
            >>> DeviceArray([0.1108716, 1.7881424], dtype=float32)

        The one standard deviation uncertainty on these parameter estimates
        (assuming the fiducial is at the maximum-likelihood estimate - which we
        know it isn't here) estimated by the square root of the inverse Fisher
        information matrix is

        .. code-block:: python

            np.sqrt(np.diag(imnn.invF))
            >>> DeviceArray([0.31980422, 0.47132865], dtype=float32)

        Note that we can compare the values estimated by the IMNN to the value
        of the mean and the variance of the target data itself, which is what
        the IMNN should be summarising

        .. code-block:: python

            np.mean(target_data)
            >>> DeviceArray(0.10693721, dtype=float32)

            np.var(target_data)
            >>> DeviceArray(1.70872, dtype=float32)

        Note that batches of data can be summarised at once using
        ``get_estimate``. In this example we will draw 10 different values of μ
        from between :math:`-10 < \\mu < 10` and 10 different values of Σ from
        between :math:`0 < \\Sigma < 10` and generate a batch of 10 different
        input data which we can summarise using the IMNN.

        .. code-block:: python

            rng, mean_keys, var_keys = jax.random.split(rng, num=3)

            mean_vals = jax.random.uniform(
                mean_keys, minval=-10, maxval=10, shape=(10,))
            var_vals = jax.random.uniform(
                var_keys, minval=0, maxval=10, shape=(10,))

            np.stack([mean_vals, var_vals], -1)
            >>> DeviceArray([[ 3.8727236,  1.6727388],
                             [-3.1113386,  8.14554  ],
                             [ 9.87299  ,  1.4134324],
                             [ 4.4837523,  1.5812075],
                             [-9.398947 ,  3.5737753],
                             [-2.0789695,  9.978279 ],
                             [-6.2622285,  6.828809 ],
                             [ 4.6470118,  6.0823894],
                             [ 5.7369494,  8.856505 ],
                             [ 4.248898 ,  5.114669 ]], dtype=float32)

            batch_target_keys = np.array(jax.random.split(rng, num=10))

            batch_target_data = jax.vmap(model_simulator)(
                batch_target_keys, (mean_vals, var_vals))

            imnn.get_estimate(batch_target_data)
            >>> DeviceArray([[ 4.6041985,  8.344688 ],
                             [-3.5172062,  7.7219954],
                             [13.229679 , 23.668312 ],
                             [ 5.745726 , 10.020965 ],
                             [-9.734651 , 21.076218 ],
                             [-1.8083427,  6.1901293],
                             [-8.626409 , 18.894459 ],
                             [ 5.7684307,  9.482665 ],
                             [ 6.7861238, 14.128591 ],
                             [ 4.900367 ,  9.472563 ]], dtype=float32)

        Parameters
        ----------
        d : float(None, input_shape)
            Input data to be compressed to score compressed parameter estimates

        Returns
        -------
        float(None, n_params):
            Score compressed parameter estimates

        Methods
        -------
        single_element:
            Returns a single score compressed summary
        multiple_elements:
            Returns a batch of score compressed summaries

        Raises
        ------
        ValueError
            If the Fisher statistics are not set after running ``fit`` or
            ``set_F_statistics``.

        Todo
        ----
        - Do proper checking on input shape (should just be a call to
          ``_check_input``)
        """
        @jax.jit
        def single_element(d):
            """Returns a single score compressed summary

            Parameters
            ----------
            d : float(input_shape)
                Input data to be compressed to score compressed parameter
                estimate

            Returns
            -------
            float(n_params,):
                Score compressed parameter estimate
            """
            return self.θ_fid + np.einsum(
                "ij,kj,kl,l->i",
                self.invF,
                self.dμ_dθ,
                self.invC,
                self.model(self.w, d) - self.μ)

        @jax.jit
        def multiple_elements(d):
            """Returns a batch of score compressed summaries

            Parameters
            ----------
            d : float(None, input_shape)
                Input data to be compressed to score compressed parameter
                estimates

            Returns
            -------
            float(None, n_params):
                Score compressed parameter estimates

            Methods
            -------
            fn:
                Returns the output of the evaluated model
            """
            def fn(d):
                """Returns the output of the evaluated model

                Parameters
                ----------
                d : float(input_shape)
                    Input data to the neural network

                Returns
                -------
                float(None, n_summaries):
                    Neural network output
                """
                return self.model(self.w, d)
            return self.θ_fid + np.einsum(
                "ij,kj,kl,ml->mi",
                self.invF,
                self.dμ_dθ,
                self.invC,
                jax.vmap(fn)(d) - self.μ)

        _check_statistics_set(self.invF, self.dμ_dθ, self.invC, self.μ)
        if len(d.shape) == 1:
            return single_element(d)
        else:
            return multiple_elements(d)

    def _setup_plot(self, ax=None, expected_detF=None, figsize=(5, 15)):
        """Builds axes for history plot

        Parameters
        ----------
        ax : mpl.axes or None, default=None
            An axes object of predefined axes to be labelled
        expected_detF : float or None, default=None
            Value of the expected determinant of the Fisher information to plot
            a horizontal line at to check fitting progress
        figsize : tuple, default=(5, 15)
            The size of the figure to be produced

        Returns
        -------
        mpl.axes:
            An axes object of labelled axes
        """
        if ax is None:
            fig, ax = plt.subplots(3, 1, sharex=True, figsize=figsize)
            plt.subplots_adjust(hspace=0.05)
        ax = [x for x in ax] + [ax[2].twinx()]
        if expected_detF is not None:
            ax[0].axhline(expected_detF, linestyle="dashed", color="black")
        ax[0].set_ylabel(r"$|{\bf F}|$")
        ax[1].axhline(1, linestyle="dashed", color="black")
        ax[1].set_ylabel(r"$|{\bf C}|$ and $|{\bf C}^{-1}|$")
        ax[1].set_yscale("log")
        ax[2].set_xlabel("Number of iterations")
        ax[2].set_ylabel(r"$\Lambda_2$")
        ax[3].set_ylabel(r"$r$")
        return ax

    def plot(self, ax=None, expected_detF=None, colour="C0", figsize=(5, 15),
             label="", filename=None, ncol=1):
        """Plot fitting history

        Plots a three panel vertical plot with the determinant of the Fisher
        information matrix in the first sublot, the covariance and the inverse
        covariance in the second and the regularisation term and the
        regularisation coupling strength in the final subplot.

        A predefined axes can be passed to fill, and these axes can be
        decorated via a call to ``_setup_plot`` (for horizonal plots for
        example).

        Example
        -------

        Assuming that an IMNN has been fit (as in the example in
        :py:meth:`imnn.imnn._imnn.IMNN.fit`) then we can make a training plot
        of the history by simply running

        .. code-block::

            imnn.fit(expected_detF=50, filename="history_plot.png")

        .. image:: /_images/history_plot.png

        Note we know the analytic value of the determinant of the Fisher
        information for this problem (:math:`|\\bf{F}|=50`) so we can add this
        line to the plot too, and save the output as a png named
        ``history_plot``.

        Parameters
        ----------
        ax : mpl.axes or None, default=None
            An axes object of predefined axes to be labelled
        expected_detF : float or None, default=None
            Value of the expected determinant of the Fisher information to plot
            a horizontal line at to check fitting progress
        colour : str or rgb/a value or list, default="C0"
            Colour to plot the lines
        figsize : tuple, default=(5, 15)
            The size of the figure to be produced
        label : str, default=""
            Name to add to description in legend
        filename : str or None, default=None
            Filename to save plot to
        ncol : int, default=1
            Number of columns to have in the legend

        Returns
        -------
        mpl.axes:
            An axes object of the filled plot
        """
        if ax is None:
            ax = self._setup_plot(expected_detF=expected_detF, figsize=figsize)
        ax[0].set_xlim(
            0, max(self.history["detF"].shape[0] - 1, ax[0].get_xlim()[-1]))
        ax[0].plot(self.history["detF"], color=colour,
                   label=r"{} $|F|$ (training)".format(label))
        ax[1].set_xlim(
            0, max(self.history["detF"].shape[0] - 1, ax[0].get_xlim()[-1]))
        ax[1].plot(self.history["detC"], color=colour,
                   label=r"{} $|C|$ (training)".format(label))
        ax[1].plot(self.history["detinvC"], linestyle="dotted", color=colour,
                   label=label + r" $|C^{-1}|$ (training)")
        ax[3].set_xlim(
            0, max(self.history["detF"].shape[0] - 1, ax[0].get_xlim()[-1]))
        ax[2].plot(self.history["Λ2"], color=colour,
                   label=r"{} $\Lambda_2$ (training)".format(label))
        ax[3].plot(self.history["r"], color=colour, linestyle="dashed",
                   label=r"{} $r$ (training)".format(label))
        if self.validate:
            ax[0].plot(self.history["val_detF"], color=colour,
                       label=r"{} $|F|$ (validation)".format(label),
                       linestyle="dotted")
            ax[1].plot(self.history["val_detC"], color=colour,
                       label=r"{} $|C|$ (validation)".format(label),
                       linestyle="dotted")
            ax[1].plot(self.history["val_detinvC"],
                       color=colour,
                       label=label + r" $|C^{-1}|$ (validation)",
                       linestyle="dashdot")
            ax[2].plot(self.history["val_Λ2"], color=colour,
                       label=r"{} $\Lambda_2$ (validation)".format(label),
                       linestyle="dotted")
            ax[3].plot(self.history["val_r"], color=colour,
                       label=r"{} $r$ (validation)".format(label),
                       linestyle="dashdot")
        h1, l1 = ax[2].get_legend_handles_labels()
        h2, l2 = ax[3].get_legend_handles_labels()
        ax[0].legend(bbox_to_anchor=(1.0, 1.0), frameon=False, ncol=ncol)
        ax[1].legend(frameon=False, bbox_to_anchor=(1.0, 1.0), ncol=ncol * 2)
        ax[3].legend(h1 + h2, l1 + l2, bbox_to_anchor=(1.05, 1.0),
                     frameon=False, ncol=ncol * 2)

        if filename is not None:
            plt.savefig(filename, bbox_inches="tight", transparent=True)
        return ax



import setuptools
import io

with io.open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="IMNN",
    version="0.3.2",
    author="Tom Charnock",
    author_email="charnock@iap.fr",
    description="Using neural networks to extract sufficient statistics from \
        data by maximising the Fisher information",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://bitbucket.org/tomcharnock/imnn.git",
    packages=["imnn", "imnn.imnn", "imnn.lfi", "imnn.utils"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "jax>=0.2.24",
        "jaxlib>=0.1.73",
        "tensorflow",
        "tqdm>=4.48.2",
        "tensorflow-probability[jax]",
        "numpy",
        "scipy",
        "matplotlib"],
    include_package_data=True
)



import tqdm.auto as tqdm
import jax
from jax.experimental.host_callback import id_tap
import jax.numpy as np


def progress_bar(max_iterations, patience, print_rate):

    if print_rate is not None:
        if max_iterations < 10000:
            pbar = tqdm.tqdm(total=max_iterations)
        else:
            pbar = tqdm.tqdm()
        remainder = max_iterations % print_rate

    def _update_pbar(args, transform):
        updates, counter, patience_counter, detF, detC, detinvC, Λ2, r = args
        pbar.update(updates)
        pbar.set_postfix(
            patience=patience_counter,
            detF=detF[counter - 1],
            detC=detC[counter - 1],
            detinvC=detinvC[counter - 1],
            Λ2=Λ2[counter - 1],
            r=r[counter - 1])

    def _close_pbar(args, transform):
        max_detF, best_w, detF, detC, detinvC, Λ2, r, \
            counter, patience_counter, state, rng = args
        updates = counter - pbar.n - 1
        pbar.update(updates)
        pbar.set_postfix(
            max_detF=max_detF,
            final_detF=detF[counter - 1],
            final_detC=detC[counter - 1],
            final_detinvC=detinvC[counter - 1],
            final_Λ2=Λ2[counter - 1],
            final_r=r[counter - 1],
            patience_reached=patience_counter)
        pbar.close()
        return False

    def _fit(func):
        def fn(x):
            return func(x)
        return fn

    def _update_controller(args):
        counter = args[0]
        jax.lax.cond(
            np.logical_and(
                np.logical_and(
                    np.greater(counter, 0),
                    np.equal(counter % print_rate, 0)),
                np.not_equal(counter, max_iterations - remainder)),
            lambda _: id_tap(
                _update_pbar, (print_rate,) + args),
            lambda _: (print_rate,) + args,
            operand=None)

        jax.lax.cond(
            np.equal(counter, max_iterations - remainder),
            lambda _: id_tap(
                _update_pbar, (remainder,) + args),
            lambda _: (remainder,) + args,
            operand=None)

    def _close_controller(cond, inputs):
        jax.lax.cond(
            cond,
            lambda _: inputs,
            lambda _: id_tap(
                _close_pbar, inputs),
            operand=None)
        return cond

    def _progress_bar(func):
        def fn(inputs):
            max_detF, best_w, detF, detC, detinvC, Λ2, r, \
                counter, patience_counter, state, rng = inputs
            _update_controller(
                (counter, patience_counter, detF, detC, detinvC, Λ2, r))
            return _close_controller(func(inputs), inputs)
        return fn

    if print_rate is not None:
        return _progress_bar
    else:
        return _fit



import matplotlib.pyplot as plt
import jax.numpy as np


class LikelihoodFreeInference:
    """Base class (some functionality) for likelihood free inference methods

    Mostly used for plotting

    Parameters
    ----------
    gridsize : list
        The number of grid points to evaluate the marginal distribution on for
        each parameter
    ranges : list
        A list of arrays containing the gridpoints for the marginal
        distribution for each parameter
    marginal : list of lists
        A list of rows and columns of marginal distributions of a corner plot

    Methods
    -------
    prior
        A prior distribution which can be evaluated and sampled from (should
        also contain a ``low`` and a ``high`` attribute with appropriate
        ranges)

    Todo
    ----
    pytests need writing
    """
    def __init__(self, prior, gridsize=100, marginals=None):
        """Constructor method

        Parameters
        ----------
        prior: fn
            A prior distribution which can be evaluated and sampled from
            (should also contain a ``low`` and a ``high`` attribute with
            appropriate ranges)
        gridsize : int or list, default=100
            The number of grid points to evaluate the marginal distribution on
            for every parameter (int) or each parameter (list)
        marginals : float(n_targets, gridsize*n_params) or None, default=None
            The full distribution evaluated on a grid to put into marginal list
        """
        self.prior = prior
        try:
            self.n_params = self.prior.event_shape[0]
        except Exception:
            raise ValueError(
                "`prior` has no event_shape - this should be `n_params`")
        if not hasattr(self.prior, "low"):
            raise ValueError(
                "`prior` must have (or be given by assignment) a `low` " +
                "attribute describing the minimum allowed value for each " +
                "parameter value")
        if not hasattr(self.prior, "low"):
            raise ValueError(
                "`prior` must have (or be given by assignment) a `low` " +
                "attribute describing the minimum allowed value for each " +
                "parameter value")
        self.gridsize = self.get_gridsize(gridsize, self.n_params)
        self.ranges = [
            np.linspace(
                self.prior.low[i],
                self.prior.high[i],
                self.gridsize[i])
            for i in range(self.n_params)]
        self.marginals = self.put_marginals(marginals)

    def get_gridsize(self, gridsize, size):
        """Propogates gridpoints per parameter into a list if int provided

        Parameters
        ----------
        gridsize : int or list
            The number of grid points to evaluate the marginal distribution on
            for every parameter (int) or each parameter (list)
        size : int
            Number of parameters describing size of gridsize list

        Returns
        -------
        list:
            The number of gridpoints to evaluate marginals for each parameter

        Raises
        ------
        ValueError
            If list passed is wrong length
        TypeError
            If gridsize is not int or list of correct length
        """
        if isinstance(gridsize, int):
            gridsize = [gridsize for i in range(size)]
        elif isinstance(gridsize, list):
            if len(gridsize) == size:
                gridsize = gridsize
            else:
                raise ValueError(
                    f"`gridsize` is a list of length {len(gridsize)} but " +
                    f"`shape` determined by `input` is {size}")
        else:
            raise TypeError("`gridsize` is not a list or an integer")
        return gridsize

    def get_levels(self, marginal, ranges, levels=[0.68, 0.95]):
        """ Used for approximately calculating confidence region isocontours

        To calculate the values of the marginal distribution whose isocontour
        contains approximately `levels` specified fraction of samples drawn
        from the distribution the marginal distribution is sorted by value and
        normalised to one. If the distribution does is significantly non-zero
        outside of the range then this will cause large biases! The cumulative
        sum of the sorted values are then calculated and the value at which
        the index of this normalised cumulative distribution is closest to the
        desired level is used to return the value of the original marginal
        distribution.

        Parameters
        ----------
        marginal : float(gridsize*n_param)
            The gridded marginal distribution to find the isocontour values of
        ranges : list
            List of the grid points for each parameter
        levels : list, default=[0.68, 0.95]
            The fraction describing the percentage of samples inside the
            isocontour

        Returns
        -------
        list:
            The values of the isocontours of the marginal distributions
        """
        domain_volume = 1
        for i in range(len(ranges)):
            domain_volume *= ranges[i][1] - ranges[i][0]
        sorted_marginal = np.sort(marginal.flatten())[::-1]
        cdf = np.cumsum(sorted_marginal / sorted_marginal.sum())
        value = []
        for level in levels[::-1]:
            this_value = sorted_marginal[np.argmin(np.abs(cdf - level))]
            if len(value) == 0:
                value.append(this_value)
            elif this_value <= value[-1]:
                break
            else:
                value.append(this_value)
        return value

    def setup_plot(self, ax=None, ranges=None, axis_labels=None,
                   figsize=(10, 10), format=False):
        """Builds corner plot

        ax : list of matplotlib axes or None, default=None
            If ax is None a new figure and axes are created to make a corner
            plot otherwise a set of axes are formatted correctly for a corner
            plot. If existing ax is not correctly shaped there will be an error
        ranges : list, default=None
            The list of ranges to set the number of rows and columns (if this
            is None there will be an error)
        axis_labels : list of str or None, default=None
            A list of names for each parameter, no axis labels if None
        figsize : tuple, default=(10, 10)
            The size (in inches) to create a figure (if ax is None)
        format : bool, default=False
            If formatting is not needed

        Returns
        -------
        axes object:
            The formatted matplotlib axes
        """
        rows = len(ranges)
        columns = len(ranges)
        if ax is None:
            fig, ax = plt.subplots(rows, columns, figsize=figsize)
            plt.subplots_adjust(wspace=0.1, hspace=0.1)
        elif not format:
            return ax
        for column in range(columns):
            for row in range(rows):
                if column > row:
                    ax[row, column].axis("off")
                else:
                    if row > 0:
                        ax[row, column].set_ylim(
                            ranges[row][0],
                            ranges[row][-1])
                        if (column == 0) and (axis_labels is not None):
                            ax[row, column].set_ylabel(axis_labels[row])
                    else:
                        ax[row, column].set_yticks([])
                    if row < rows - 1:
                        ax[row, column].set_xticks([])
                    if column > 0:
                        ax[row, column].set_yticks([])
                    if column < columns - 1:
                        ax[row, column].set_xlim(
                            ranges[column][0],
                            ranges[column][-1])
                        if (row == rows - 1) and (axis_labels is not None):
                            ax[row, column].set_xlabel(axis_labels[column])
                    else:
                        ax[row, column].set_xticks([])
        return ax

    def _scatter_plot(self, ax=None, ranges=None, points=None, label=None,
                      axis_labels=None, colours=None, hist=True, s=5.,
                      alpha=1., figsize=(10, 10), linestyle="solid",
                      target=None, format=False, ncol=2,
                      bbox_to_anchor=(0.0, 1.0)):
        """Plotter for scatter plots

        Plots scatter plots for points (parameters or summaries) in 2D subplots
        and the histogram of points in the 1D diagonal subplots.

        Parameters
        ----------
        ax : list of matplotlib axes or None, default=None
            If ax is None a new figure and axes are created to make a corner
            plot otherwise a set of axes are formatted correctly for a corner
            plot. If existing ax is not correctly shaped there will be an error
        ranges : list, default=None
            The list of ranges to set the number of rows and columns (if this
            is None the ranges set on initialisation will be used)
        points : float(n_targets, n_points, {n_params} or {n_summaries})
            The points to scatter plot
        label : str or None, default=None
            Name to be used in the legend
        axis_labels : list of str or None, default=None
            A list of names for each parameter, no axis labels if None
        colours : str or list or None, default=None
            The colour or list of colours to use for the different targets, if
            None then the normal matplotlib colours are used
        hist : bool, default=True
            Whether or not to plot histograms on the diagonal of the corner
            plot
        s : float, default=5.
            The size of the marker points in the scatter plot
        alpha : float, default=1.
            The amount of alpha colour for the marker points
        figsize : tuple, default=(10, 10)
            The size (in inches) to create a figure (if ax is None)
        linestyle : str, default="solid"
            Linestyle for the histograms
        target : None or int or list, default=None
            The indices to choose which target's points are plotted
        format : bool, default=False
            If formatting is not needed
        ncols : int, default=2
            Number of columns for the legend
        bbox_to_anchor : tuple, default=(0.0, 1.0)
            Position to fix the legend to

        Returns
        -------
        axes object:
            The scatter plot axes

        Raises
        ------
        ValueError
            if colour input is not correct

        Todo
        ----
        There are many extra matplotlib parameters which could be passed,
        although this is not done because the axis is returned which can then
        be manipulated.
        """
        if ranges is None:
            raise ValueError("`ranges` must be provided")
        if points is None:
            raise ValueError("`points` to scatter must be provided")
        targets, n_targets = self.target_choice(target)
        if colours is None:
            colours = [f"C{i}" for i in range(n_targets)]
        elif isinstance(colours, str):
            colours = [colours for i in range(n_targets)]
        elif isinstance(colours, list):
            pass
        else:
            raise ValueError(
                "`colours` must be None, a color as a string or a list of " +
                "colours")
        if ranges is None:
            ranges = self.ranges
        if label is None:
            label = ""
        rows = len(ranges)
        columns = len(ranges)
        ax = self.setup_plot(ax=ax, ranges=ranges, axis_labels=axis_labels,
                             figsize=figsize, format=format)
        for column in range(columns):
            for row in range(rows):
                for i, target in enumerate(targets):
                    if (column == row) and hist:
                        if column < columns - 1:
                            ax[row, column].hist(
                                points[target][:, row],
                                bins=ranges[row],
                                color=colours[i],
                                linestyle=linestyle,
                                density=True,
                                histtype="step")
                        else:
                            ax[row, column].hist(
                                points[target][:, row],
                                bins=ranges[column],
                                color=colours[i],
                                linestyle=linestyle,
                                density=True,
                                histtype="step",
                                orientation="horizontal")

                    elif column < row:
                        ax[row, column].scatter(
                            points[target][:, column],
                            points[target][:, row],
                            s=s,
                            color=colours[i],
                            alpha=alpha,
                            label=label + f" Target {target+1}")
                        h, lab = ax[row, column].get_legend_handles_labels()
        if label != "":
            ax[0, 0].legend(h, lab, frameon=False,
                            bbox_to_anchor=bbox_to_anchor, ncol=ncol)
        return ax

    def marginal_plot(self, ax=None, ranges=None, marginals=None, known=None,
                      label=None, axis_labels=None, levels=None,
                      linestyle="solid", colours=None, target=None,
                      format=False, ncol=2, bbox_to_anchor=(1.0, 1.0)):
        """Plots the marginal distribution corner plots

        Plots the 68% and 95% (approximate) confidence contours for each 2D
        parameter pair and the 1D densities as a corner plot.

        Parameters
        ----------
        ax : list of matplotlib axes or None, default=None
            If ax is None a new figure and axes are created to make a corner
            plot otherwise a set of axes are formatted correctly for a corner
            plot. If existing ax is not correctly shaped there will be an error
        ranges : list, default=None
            The list of ranges to set the number of rows and columns (if this
            is None the ranges set on initialisation will be used)
        marginals : list of lists (n_params * n_params) or None, default=None
            The marginal distributions for every parameter and every 2D
            combination. If None the marginals defined in the class are used
            (if they exist)
        known : float(n_params,)
            Values to plot known parameter value lines at
        label : str or None, default=None
            Name to be used in the legend
        axis_labels : list of str or None, default=None
            A list of names for each parameter, no axis labels if None
        linestyle : str, default="solid"
            Linestyle for the histograms
        colours : str or list or None, default=None
            The colour or list of colours to use for the different targets, if
            None then the normal matplotlib colours are used
        target : None or int or list, default=None
            The indices to choose which target's points are plotted
        format : bool, default=False
            If formatting is not needed
        ncols : int, default=2
            Number of columns for the legend
        bbox_to_anchor : tuple, default=(0.0, 1.0)
            Position to fix the legend to

        Returns
        -------
        axes object:
            The scatter plot axes

        Raises
        ------
        ValueError
            if `marginals` is not provided
        TypeError
            if `marginals` is not a list

        Todo
        ----
        There are many extra matplotlib parameters which could be passed,
        although this is not done because the axis is returned which can then
        be manipulated.
        """
        targets, n_targets = self.target_choice(target)
        if (marginals is None) and (self.marginals is None):
            raise ValueError(
                "Need to provide `marginal` or run `get_marginals()`")
        elif marginals is None:
            marginals = self.marginals
        elif not isinstance(marginals, list):
            raise TypeError
        if levels is None:
            levels = [0.68, 0.95]
        if colours is None:
            colours = [f"C{i}" for i in range(n_targets)]
        if type(colours) is str:
            colours = [colours for i in range(n_targets)]
        if ranges is None:
            ranges = self.ranges
        if label is None:
            label = ""
        rows = len(ranges)
        columns = len(ranges)
        ax = self.setup_plot(ax=ax, ranges=ranges, axis_labels=axis_labels,
                             format=format)
        for column in range(columns):
            for row in range(rows):
                for i, target in enumerate(targets):
                    if column == row:
                        if column < columns - 1:
                            ax[row, column].plot(
                                ranges[row],
                                marginals[row][column][target],
                                color=colours[i],
                                linestyle=linestyle,
                                label=label)
                            if known is not None:
                                ax[row, column].axvline(
                                    known[column],
                                    color="black",
                                    linestyle="dashed")
                        else:
                            ax[row, column].plot(
                                marginals[row][column][target],
                                ranges[row],
                                color=colours[i],
                                linestyle=linestyle)
                            if known is not None:
                                ax[row, column].axhline(
                                    known[column],
                                    color="black",
                                    linestyle="dashed")
                    elif column < row:
                        ax[row, column].contour(
                            ranges[column],
                            ranges[row],
                            marginals[row][column][target].T,
                            colors=colours[i],
                            linestyles=[linestyle],
                            levels=self.get_levels(
                                marginals[row][column][target],
                                [ranges[column], ranges[row]],
                                levels=levels))
                        if known is not None:
                            ax[row, column].axvline(
                                known[column],
                                color="black",
                                linestyle="dashed")
                            ax[row, column].axhline(
                                known[row],
                                color="black",
                                linestyle="dashed")
        if label != "":
            ax[0, 0].legend(
                frameon=False,
                bbox_to_anchor=bbox_to_anchor,
                ncol=ncol)
        return ax

    def put_marginals(self, marginals):
        """Creates list of 1D and 2D marginal distributions ready for plotting

        The marginal distribution lists from full distribution array. For every
        parameter the full distribution is summed over every other parameter to
        get the 1D marginals and for every combination the 2D marginals are
        calculated by summing over the remaining parameters. The list is made
        up of a list of n_params lists which contain n_columns number of
        objects.

        Parameters
        ----------
        marginals : float(n_targets, gridsize*n_params)
            The full distribution from which to calculate the marginal
            distributions

        Returns
        -------
        list of lists:
            The 1D and 2D marginal distributions for each parameter (of pair)
        """
        if marginals is None:
            return None
        _marginals = []
        for row in range(self.n_params):
            _marginals.append([])
            for column in range(self.n_params):
                if column == row:
                    _marginals[row].append(
                        marginals.sum(
                            tuple(i + 1 for i in range(self.n_params)
                                  if i != row)))
                    _marginals[row][-1] /= np.expand_dims(
                        np.sum(_marginals[row][-1], 1)
                        * (self.ranges[row][1] - self.ranges[row][0]),
                        1)
                elif column < row:
                    _marginals[row].append(
                        marginals.sum(
                            tuple(i + 1 for i in range(self.n_params)
                                  if (i != row) and (i != column))))
        self.marginals = _marginals
        return _marginals

    def target_choice(self, target):
        """Returns the indices of targets to plot and the number of targets

        Parameters
        ----------
        target : None or int or list
            The indices of the targets to plot. If None then ``n_targets`` from
            the class instance is used

        Returns
        -------
        list:
            The indices of the targets to be plotted
        int:
            The number of targets to be plotted
        """
        if target is None:
            if self.n_targets is None:
                raise ValueError("``n_targets`` class attribute not defined")
            targets = range(self.n_targets)
            n_targets = self.n_targets
        elif isinstance(target, list):
            targets = target
            n_targets = len(target)
        elif isinstance(targets, int):
            targets = [target]
            n_targets = 1
        else:
            raise ValueError(
                "`target` must be None (to use targets defined in " +
                "initialisation), a list of indices for the desired targets " +
                "to be plotted or an integer for the single target to be " +
                "plotted.")
        return targets, n_targets



import jax
import jax.numpy as np
from functools import partial
from imnn.imnn._imnn import _IMNN
from imnn.utils import value_and_jacrev
from imnn.utils.utils import _check_simulator


class SimulatorIMNN(_IMNN):
    """Information maximising neural network fit with simulations on-the-fly

    Defines the function to get simulations and compress them using an XLA
    compilable simulator.

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    random number generators are generated and used to generate a set of
    :math:`n_s` simulations,
    :math:`{\\bf d}^i={\\rm simulator}({\\rm seed}^i, \\theta^\\rm{fid})` at
    the fiducial model parameters, :math:`\\theta^\\rm{fid}`, and these are
    passed direrectly through a network :math:`f_{{\\bf w}}({\\bf d})` with
    network parameters :math:`{\\bf w}` to obtain network outputs
    :math:`{\\bf x}^i` and autodifferentiation is used to get the derivative of
    :math:`n_d` of these outputs with respect to the physical model parameters,
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha`, where
    :math:`\\alpha` labels the physical parameter. With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    Once the loss function is calculated the automatic gradient is then
    calculated and used to update the network parameters via the optimiser
    function.

    Methods
    -------
    simulator:
        A function for generating a simulation on-the-fly (XLA compilable)

    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, simulator):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also checks
        validity of simulator and sets the ``simulate`` attribute to ``True``.

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        simulator : fn
            A function that generates a single simulation from a random number
            generator and a tuple (or array) of parameter values at which to
            generate the simulations. For the purposes of use in LFI/ABC
            afterwards it is also useful for the simulator to be able to
            broadcast to a batch of simulations on the zeroth axis
            ``fn(int(2,), float([None], n_params)) ->
            float([None], input_shape)``
        """
        super().__init__(
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser)
        self.simulator = _check_simulator(simulator)
        self.simulate = True

    def _get_fitting_keys(self, rng):
        """Generates random numbers for simulation

        Parameters
        ----------
        rng : int(2,)
            A random number generator

        Returns
        -------
        int(2,), int(2,), int(2,)
            A new random number generator and random number generators for
            fitting (and validation)
        """
        return jax.random.split(rng, num=3)

    def get_summaries(self, w, key, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        A random seed for each simulation is obtained and ``n_d`` of them are
        used to calculate the network outputs of each of these simulations as
        well as the derivative of these network outputs with respect to the
        model parameters as calculated using jax autodifferentiation. The
        remaining ``n_s - n_d`` network outputs are then calculated and
        concatenated to those already calculated.

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=False
            Whether to get summaries of the validation set

        Returns
        -------
        float(n_s, n_summaries):
            The set of all network outputs used to calculate the covariance
        float(n_d, n_summaries, n_params):
            The set of all network output derivatives wrt model parameters

        Methods
        -------
        get_summary:
            Return a single network output
        get_derivatives:
            Return the Jacobian of the network outputs wrt model parameters
        """
        def get_summary(key, θ):
            """Return a single network output

            Parameters
            ----------
            key : int(2,)
                A random number generator for generating simulations
            θ : float(n_params,)
                The value of the model parameters to generate the simulation at

            Returns
            -------
            float(n_summaries):
                A single simulation passed through the neural network
            """
            return self.model(w, self.simulator(key, θ))

        def get_derivatives(key):
            """Return the Jacobian of the network outputs wrt model parameters

            Parameters
            ----------
            key : int(2,)
                A random number generator for generating simulations

            Returns
            -------
            float(n_summaries):
                A single simulation passed through the neural network
            float(n_summaries, n_params)
                The derivative of the network output wrt model parameters
            """
            return value_and_jacrev(get_summary, argnums=1)(key, self.θ_fid)

        keys = np.array(jax.random.split(key, num=self.n_s))
        summaries, derivatives = jax.vmap(get_derivatives)(keys[:self.n_d])
        if self.n_s > self.n_d:
            summaries = np.vstack([
                summaries,
                jax.vmap(partial(get_summary, θ=self.θ_fid))(keys[self.n_d:])])
        return summaries, derivatives



import jax
import jax.numpy as np
from jax.scipy.stats import norm, multivariate_normal
from imnn.lfi import LikelihoodFreeInference


class GaussianApproximation(LikelihoodFreeInference):
    """Uses Fisher information and parameter estimates approximate marginals

    Since the inverse of the Fisher information matrix describes the minimum
    variance of some estimator we can use it to make an approximate (Gaussian-
    distributed) estimate of the target distribution. Note that this will not
    reflect the true shape of the target distribution and as likely to
    underestimate the distribution as overestimate it. Furthermore, if the
    Fisher information matrix is calculated far from the estimate of the
    parameter values then its value may not be representative of the Fisher
    information at that position and so the variance estimated from its inverse
    be incorrect.

    Parameters
    ----------
    parameter_estimates: float(n_targets, n_params)
        The parameter estimates of each target data
    invF: float(n_targets, n_params, n_params)
        The inverse Fisher information matrix for each target
    marginals: list of lists
        The 1D and 2D marginal distribution for each target

    Todo
    ----
    type checking and pytests need implementing
    """
    def __init__(self, parameter_estimates, invF, prior, gridsize=100):
        """Constructor method

        Parameters
        ----------
        parameter_estimates: float(n_targets, n_params)
            The parameter estimates of each target data
        invF: float(n_targets, n_params, n_params)
            The inverse Fisher information matrix for each target
        prior: fn
            A prior distribution which can be evaluated and sampled from
            (should also contain a ``low`` and a ``high`` attribute with
            appropriate ranges)
        gridsize : int or list, default=100
            The number of grid points to evaluate the marginal distribution on
            for every parameter (int) or each parameter (list)
        """
        super().__init__(
            prior=prior,
            gridsize=gridsize)
        if len(parameter_estimates.shape) == 0:
            parameter_estimates = np.expand_dims(parameter_estimates, 0)
        if len(parameter_estimates.shape) == 1:
            parameter_estimates = np.expand_dims(parameter_estimates, 0)
        self.parameter_estimates = parameter_estimates
        self.n_targets = self.parameter_estimates.shape[0]
        self.n_params = self.parameter_estimates.shape[-1]
        self.invF = invF
        self.marginals = self.get_marginals()

    def get_marginals(self, parameter_estimates=None, invF=None, ranges=None,
                      gridsize=None):
        """
        Creates list of 1D and 2D marginal distributions ready for plotting

        The marginal distribution lists from full distribution array. For every
        parameter the full distribution is summed over every other parameter to
        get the 1D marginals and for every combination the 2D marginals are
        calculated by summing over the remaining parameters. The list is made
        up of a list of n_params lists which contain n_columns number of
        objects. The value of the distribution comes from

        Parameters
        ----------
        parameter_estimates: float(n_targets, n_params) or None, default=None
            The parameter estimates of each target data. If None the class
            instance parameter estimates are used
        invF: float(n_targets, n_params, n_params) or None, default=None
            The inverse Fisher information matrix for each target. If None the
            class instance inverse Fisher information matrices are used
        ranges : list or None, default=None
            A list of arrays containing the gridpoints for the marginal
            distribution for each parameter. If None the class instance ranges
            are used determined by the prior range
        gridsize : list or None, default=None
            If using own `ranges` then the gridsize for these ranges must be
            passed (not checked)

        Returns
        -------
        list of lists:
            The 1D and 2D marginal distributions for each parameter (of pair)

        Todo
        ----
        Need to multiply the distribution by the prior to get the posterior
        Maybe move to TensorFlow probability?
        Make sure that using several Fisher estimates works
        """
        if parameter_estimates is None:
            parameter_estimates = self.parameter_estimates
        n_targets = parameter_estimates.shape[0]
        if invF is None:
            invF = self.invF
        if ranges is None:
            ranges = self.ranges
        if gridsize is None:
            gridsize = self.gridsize
        marginals = []
        for row in range(self.n_params):
            marginals.append([])
            for column in range(self.n_params):
                if column == row:
                    marginals[row].append(
                        jax.vmap(
                            lambda mean, _invF: norm.pdf(
                                ranges[column],
                                mean,
                                np.sqrt(_invF)))(
                                    parameter_estimates[:, column],
                                    invF[:, column, column]))
                elif column < row:
                    X, Y = np.meshgrid(ranges[row], ranges[column])
                    unravelled = np.vstack([X.ravel(), Y.ravel()]).T
                    marginals[row].append(
                        jax.vmap(
                            lambda mean, _invF: multivariate_normal.pdf(
                                unravelled,
                                mean,
                                _invF).reshape(
                                    ((gridsize[column], gridsize[row]))))(
                            parameter_estimates[:, [row, column]],
                            invF[:,
                                 [row, row, column, column],
                                 [row, column, row, column]].reshape(
                                (n_targets, 2, 2))))
        return marginals
